__all__ = ('WebhookRepr',)

from scarletio import export

from .preinstanced import WebhookType
from .utils import create_partial_webhook_from_id
from .webhook_base import WebhookBase


@export
class WebhookRepr(WebhookBase):
    """
    Represents a Discord webhook's user representation.
    
    Attributes
    ----------
    avatar_hash : `int`
        The webhook representation's avatar's hash in `uint128`.
    
    avatar_type : ``IconType``
        The webhook representation's avatar's type.
    
    channel_id : `int`
        The channel, where the webhook is going to send it's messages.
    
    id : `int`
        The webhook representation's unique identifier number.
    
    name : str
        The webhook representation's username.
    
    type : ``WebhookType``
        The webhook's type.
    
    Notes
    -----
    Instances of the type support weakreferencing.
    """
    __slots__ = ()
    
    @classmethod
    def from_data(cls, data, webhook_id, webhook_type, channel_id):
        """
        Creates a webhook representation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Included user data.
        
        webhook_id : `int`
            The respective webhook's identifier number.
        
        webhook_type : ``WebhookType``
            The respective webhook's type.
        
        channel_id : `int`
            The respective webhook's channel's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._update_attributes(data)
        self.channel_id = channel_id
        self.id = webhook_id
        self.type = webhook_type
        return self
    
    
    @property
    def webhook(self):
        """
        Creates a partial webhook from the webhook data included with the webhook representation.
        
        Returns
        -------
        webhook : ``Webhook``
        """
        return create_partial_webhook_from_id(self.id, '', webhook_type = self.type, channel_id = self.channel_id)



__all__ = ()

from scarletio import copy_docs

from ..packets.utils import create_rtp_data_lite, create_rtp_header_lite

from .base import EncryptionAdapterBase
from .nacl import crypto_secretbox, crypto_secretbox_KEYBYTES, crypto_secretbox_NONCEBYTES, crypto_secretbox_open


class EncryptionAdapter__xsalsa20_poly1305(EncryptionAdapterBase):
    """
    `xsalsa20_poly1305` encryption adapter.
    
    Attributes
    ----------
    key : `bytes`
        Secret key used for encrypting / decrypting.
    """
    __slots__ = ()
    
    available = True if crypto_secretbox_KEYBYTES else False
    key_length = crypto_secretbox_KEYBYTES
    name = 'xsalsa20_poly1305'
    nonce_length = crypto_secretbox_NONCEBYTES
    priority = 1
    
    
    @copy_docs(EncryptionAdapterBase.__new__)
    def __new__(cls, key):
        self = object.__new__(cls)
        self.key = key
        return self
    
    
    @copy_docs(EncryptionAdapterBase.create_send_packet)
    def create_send_packet(self, voice_client, plain_data):
        header = create_rtp_header_lite(
            voice_client._sequence, voice_client._timestamp, voice_client._audio_source
        )
        nonce = header.ljust(self.nonce_length, b'\x00')
        payload = crypto_secretbox(plain_data, nonce, self.key)
        return create_rtp_data_lite(header, payload, None)
    
    
    @copy_docs(EncryptionAdapterBase.process_received_payload)
    def process_received_payload(self, rtp_packet):
        nonce = bytes(rtp_packet.header).ljust(self.nonce_length, b'\x00')
        return crypto_secretbox_open(bytes(rtp_packet.payload), nonce, self.key)



__all__ = ()

from scarletio import set_docs

from ....env import CACHE_PRESENCE

from ...channel import Channel
from ...embedded_activity import EmbeddedActivity
from ...emoji import Emoji
from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, flag_parser_factory, force_string_parser_factory, int_parser_factory,
    negated_bool_parser_factory, nullable_entity_parser_factory, nullable_string_parser_factory,
    preinstanced_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_dictionary_putter_factory, entity_id_optional_putter_factory,
    entity_id_putter_factory, flag_putter_factory, force_string_putter_factory, int_optional_putter_factory,
    int_putter_factory, negated_bool_optional_putter_factory, nullable_entity_putter_factory,
    nullable_string_optional_putter_factory, preinstanced_array_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_dictionary_validator_factory, entity_id_validator_factory, flag_validator_factory,
    force_string_validator_factory, int_conditional_validator_factory, int_options_validator_factory,
    nullable_entity_dictionary_validator_factory, nullable_entity_set_validator_factory,
    nullable_entity_validator_factory, nullable_string_validator_factory, preinstanced_array_validator_factory,
    preinstanced_validator_factory
)
from ...localization import Locale
from ...localization.utils import LOCALE_DEFAULT
from ...role import Role
from ...scheduled_event import ScheduledEvent
from ...soundboard import SoundboardSound
from ...stage import Stage
from ...sticker import Sticker
from ...user import ClientUserBase, User, VoiceState

from ..guild_incidents import GuildIncidents
from ..guild_inventory_settings import GuildInventorySettings

from .constants import (
    AFK_TIMEOUT_DEFAULT, AFK_TIMEOUT_OPTIONS, DESCRIPTION_LENGTH_MAX, MAX_PRESENCES_DEFAULT,
    MAX_STAGE_CHANNEL_VIDEO_USERS_DEFAULT, MAX_USERS_DEFAULT, MAX_VOICE_CHANNEL_VIDEO_USERS_DEFAULT, NAME_LENGTH_MAX,
    NAME_LENGTH_MIN
)
from .guild_boost_perks import LEVELS
from .flags import SystemChannelFlag
from .preinstanced import (
    ExplicitContentFilterLevel, GuildFeature, HubType, MfaLevel, MessageNotificationLevel, NsfwLevel, VerificationLevel
)

# afk_channel_id

parse_afk_channel_id = entity_id_parser_factory('afk_channel_id')
put_afk_channel_id = entity_id_optional_putter_factory('afk_channel_id')
validate_afk_channel_id = entity_id_validator_factory('afk_channel_id', Channel)

# afk_timeout

parse_afk_timeout = int_parser_factory('afk_timeout', AFK_TIMEOUT_DEFAULT)
put_afk_timeout = int_optional_putter_factory('afk_timeout', AFK_TIMEOUT_DEFAULT)
validate_afk_timeout = int_options_validator_factory('afk_timeout', AFK_TIMEOUT_OPTIONS, 0)

# approximate_online_count

parse_approximate_online_count = int_parser_factory('approximate_presence_count', 0)
put_approximate_online_count = int_putter_factory('approximate_presence_count')
validate_approximate_online_count = int_conditional_validator_factory(
    'approximate_online_count',
    0,
    lambda approximate_online_count : approximate_online_count >= 0,
    '>= 0',
)

# approximate_user_count

parse_approximate_user_count = int_parser_factory('approximate_member_count', 0)
put_approximate_user_count = int_putter_factory('approximate_member_count')
validate_approximate_user_count = int_conditional_validator_factory(
    'approximate_user_count',
    0,
    lambda approximate_user_count : approximate_user_count >= 0,
    '>= 0',
)

# available

parse_available = negated_bool_parser_factory('unavailable', True)
put_available = negated_bool_optional_putter_factory('unavailable', True)
validate_available = bool_validator_factory('unavailable', True)


# boost_count

parse_boost_count = int_parser_factory('premium_subscription_count', 0)
put_boost_count = int_optional_putter_factory('premium_subscription_count', 0)
validate_boost_count = int_conditional_validator_factory(
    'boost_count',
    0,
    lambda boost_count : boost_count >= 0,
    '>= 0',
)


# boost_level

parse_boost_level = int_parser_factory('premium_tier', 0)
put_boost_level = int_putter_factory('premium_tier')
validate_boost_level = int_options_validator_factory('boost_level', frozenset(LEVELS.keys()), 0)


# boost_progress_bar_enabled

parse_boost_progress_bar_enabled = bool_parser_factory('premium_progress_bar_enabled', False)
put_boost_progress_bar_enabled = bool_optional_putter_factory('premium_progress_bar_enabled', False)
validate_boost_progress_bar_enabled = bool_validator_factory('boost_progress_bar_enabled', False)


# channels

def parse_channels(data, channels, guild_id = 0):
    """
    Parses the guild's channels from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    channels : `dict<int, Channel>`
        The guild's channels.
    guild_id : `int` = `0`, Optional
        The guild's identifier.
        
    Returns
    -------
    channels : `dict<int, Channel>`
        Returns `channels` parameter.
    """
    if channels:
        old_channels = [*channels.values()]
        channels.clear()
    
    channel_datas = data.get('channels', None)
    if (channel_datas is not None):
        for channel_data in channel_datas:
            channel = Channel.from_data(channel_data, None, guild_id, strong_cache = False)
            channels[channel.id] = channel
    
    return channels


put_channels = entity_dictionary_putter_factory('channels', Channel, force_include_internals = True)
validate_channels = entity_dictionary_validator_factory('channels', Channel)

# client_guild_profile

def parse_client_guild_profile__cache_presence(data, users, guild_id):
    return users


def parse_client_guild_profile__no_cache_presence(data, users, guild_id):
    guild_profile_datas = data.get('members', None)
    if (guild_profile_datas is not None):
        for guild_profile_data in guild_profile_datas:
            user = User._bypass_no_cache(guild_profile_data['user'], guild_profile_data, guild_id)
            if (user is not None):
                users[user.id] = user
    
    return users


if CACHE_PRESENCE:
    parse_client_guild_profile = parse_client_guild_profile__cache_presence
else:
    parse_client_guild_profile = parse_client_guild_profile__no_cache_presence

set_docs(
    parse_client_guild_profile,
    """
    Parses the guild's client's guild profile data.
    
    > Used when presence caching is disabled.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    users : `dict<int, ClientUserBase>`
        The guild's users.
    guild_id : `int` = `0`
        The guild's identifier.
        
    Returns
    -------
    users : `dict<int, ClientUserBase>`
        Returns `users` parameter.
    """
)

# default_message_notification_level

parse_default_message_notification_level = preinstanced_parser_factory(
    'default_message_notifications', MessageNotificationLevel, MessageNotificationLevel.all_messages
)
put_default_message_notification_level = preinstanced_putter_factory('default_message_notifications')
validate_default_message_notification_level = preinstanced_validator_factory(
    'default_message_notification_level', MessageNotificationLevel
)

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_optional_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# embedded_activities

def parse_embedded_activities(data, embedded_activities, guild_id = 0):
    """
    Parses the guild's embedded_activities from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    embedded_activities : `None | set<EmbeddedActivity>`
        The guild's embedded activities.
    guild_id : `int` = `0`, Optional
        The guild's identifier.
        
    Returns
    -------
    embedded_activities : `None | set<EmbeddedActivity>`
    """
    if (embedded_activities is not None):
        old_embedded_activities = [*embedded_activities]
        embedded_activities.clear()
    
    embedded_activity_datas = data.get('activity_instances', None)
    if (embedded_activity_datas is not None):
        for embedded_activity_data in embedded_activity_datas:
            embedded_activity = EmbeddedActivity.from_data(embedded_activity_data, guild_id)
            
            if embedded_activities is None:
                embedded_activities = set()
                
            embedded_activities.add(embedded_activity)
    
    if (embedded_activities is not None) and (not embedded_activities):
        embedded_activities = None
    
    return embedded_activities


def put_embedded_activities(embedded_activities, data, defaults):
    """
    Puts the given embedded activities into the given `data` json serializable object.
    
    Parameters
    ----------
    embedded_activities : `None | set<EmbeddedActivity>`
        Entity array.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if embedded_activities is None:
        embedded_activity_datas = []
    else:
        embedded_activity_datas = [
            embedded_activity.to_data(defaults = defaults, include_internals = True)
            for embedded_activity in embedded_activities
        ]
    
    data['activity_instances'] = embedded_activity_datas
    
    return data


validate_embedded_activities = nullable_entity_set_validator_factory(
    'embedded_activities', EmbeddedActivity
)

# emojis

def parse_emojis(data, emojis, guild_id = 0):
    """
    Parses the guild's emojis from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    emojis : `dict<int, Emoji>`
        The guild's emojis.
    guild_id : `int` = `0`, Optional
        The guild's identifier.
        
    Returns
    -------
    emojis : `dict<int, Emoji>`
        Returns `emojis` parameter.
    """
    if emojis:
        old_emojis = [*emojis.values()]
        emojis.clear()
    
    emoji_datas = data.get('emojis', None)
    if (emoji_datas is not None):
        for emoji_data in emoji_datas:
            emoji = Emoji.from_data(emoji_data, guild_id)
            emojis[emoji.id] = emoji
    
    return emojis

put_emojis = entity_dictionary_putter_factory('emojis', Emoji, force_include_internals = True)
validate_emojis = entity_dictionary_validator_factory('emojis', Emoji)

# explicit_content_filter_level

parse_explicit_content_filter_level = preinstanced_parser_factory(
    'explicit_content_filter', ExplicitContentFilterLevel, ExplicitContentFilterLevel.disabled
)
put_explicit_content_filter_level = preinstanced_putter_factory('explicit_content_filter')
validate_explicit_content_filter_level = preinstanced_validator_factory(
    'explicit_content_filter_level', ExplicitContentFilterLevel
)

# features

parse_features = preinstanced_array_parser_factory('features', GuildFeature)
put_features = preinstanced_array_putter_factory('features')
validate_features = preinstanced_array_validator_factory('features', GuildFeature)

# hub_type

parse_hub_type = preinstanced_parser_factory('hub_type', HubType, HubType.none)
put_hub_type = preinstanced_putter_factory('hub_type')
validate_hub_type = preinstanced_validator_factory('hub_type', HubType)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('guild_id')

# incidents

parse_incidents = nullable_entity_parser_factory('incidents_data', GuildIncidents)
put_incidents = nullable_entity_putter_factory('incidents_data', GuildIncidents, force_include_internals = True)
validate_incidents = nullable_entity_validator_factory('incidents', GuildIncidents)

# inventory_settings

parse_inventory_settings = nullable_entity_parser_factory('inventory_settings', GuildInventorySettings)
put_inventory_settings = nullable_entity_putter_factory('inventory_settings', GuildInventorySettings)
validate_inventory_settings = nullable_entity_validator_factory('inventory_settings', GuildInventorySettings)

# large

parse_large = bool_parser_factory('large', False)
put_large = bool_optional_putter_factory('large', False)
validate_large = bool_validator_factory('large', False)

# max_presences

parse_max_presences = int_parser_factory('max_presences', MAX_PRESENCES_DEFAULT)
put_max_presences = int_putter_factory('max_presences')
validate_max_presences = int_conditional_validator_factory(
    'max_presences',
    MAX_PRESENCES_DEFAULT,
    (lambda max_presences : max_presences >= 0),
    '>= 0',
)

# max_stage_channel_video_users

parse_max_stage_channel_video_users = int_parser_factory(
    'max_stage_video_channel_users', MAX_STAGE_CHANNEL_VIDEO_USERS_DEFAULT
)
put_max_stage_channel_video_users = int_putter_factory('max_stage_video_channel_users')
validate_max_stage_channel_video_users = int_conditional_validator_factory(
    'max_stage_channel_video_users',
    MAX_STAGE_CHANNEL_VIDEO_USERS_DEFAULT,
    (lambda max_stage_channel_video_users : max_stage_channel_video_users >= 0),
    '>= 0',
)

# max_users

parse_max_users = int_parser_factory('max_members', MAX_USERS_DEFAULT)
put_max_users = int_putter_factory('max_members')
validate_max_users = int_conditional_validator_factory(
    'max_users',
    MAX_USERS_DEFAULT,
    (lambda max_users : max_users >= 0),
    '>= 0',
)

# max_voice_channel_video_users

parse_max_voice_channel_video_users = int_parser_factory(
    'max_video_channel_users', MAX_VOICE_CHANNEL_VIDEO_USERS_DEFAULT
)
put_max_voice_channel_video_users = int_putter_factory('max_video_channel_users')
validate_max_voice_channel_video_users = int_conditional_validator_factory(
    'max_voice_channel_video_users',
    MAX_VOICE_CHANNEL_VIDEO_USERS_DEFAULT,
    (lambda max_voice_channel_video_users : max_voice_channel_video_users >= 0),
    '>= 0',
)

# mfa_level

parse_mfa_level = preinstanced_parser_factory('mfa_level', MfaLevel, MfaLevel.none)
put_mfa_level = preinstanced_putter_factory('mfa_level')
validate_mfa_level = preinstanced_validator_factory('mfa_level', MfaLevel)

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# nsfw_level

parse_nsfw_level = preinstanced_parser_factory('nsfw_level', NsfwLevel, NsfwLevel.none)
put_nsfw_level = preinstanced_putter_factory('nsfw_level')
validate_nsfw_level = preinstanced_validator_factory('nsfw_level', NsfwLevel)

# owner_id

parse_owner_id = entity_id_parser_factory('owner_id')
put_owner_id = entity_id_putter_factory('owner_id')
validate_owner_id = entity_id_validator_factory('owner_id', ClientUserBase)

# locale


def parse_locale(data):
    """
    Parses out the guild's locale from the given value.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    locale : ``Locale``
    """
    # When receiving interaction
    locale = data.get('locale', None)
    if (locale is not None):
        return Locale(locale)
    
    # When receiving guild itself
    locale = data.get('preferred_locale', None)
    if (locale is not None):
        return Locale(locale)
    
    return LOCALE_DEFAULT


def put_locale(locale, data, defaults):
    """
    Puts the given `locale` into the given `data` json serializable object.
    
    Parameters
    ----------
    locale : ``Locale``
        Locale to serialise.
    data : `dict<str, object>`
        Interaction resolved data.
    defaults : `bool`
        Whether default fields values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data['preferred_locale'] = locale.value
    data['locale'] = locale.value
    return data


validate_locale = preinstanced_validator_factory('locale', Locale)


# public_updates_channel_id

parse_public_updates_channel_id = entity_id_parser_factory('public_updates_channel_id')
put_public_updates_channel_id = entity_id_optional_putter_factory('public_updates_channel_id')
validate_public_updates_channel_id = entity_id_validator_factory('public_updates_channel_id', Channel)

# roles

def parse_roles(data, roles, guild_id = 0):
    """
    Parses the guild's roles from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    roles : `dict<int, Role>`
        The guild's roles.
    guild_id : `int` = `0`, Optional
        The guild's identifier.
        
    Returns
    -------
    roles : `dict<int, Role>`
        Returns `roles` parameter.
    """
    if roles:
        old_roles = [*roles.values()]
        roles.clear()
    
    role_datas = data.get('roles', None)
    if (role_datas is not None):
        for role_data in role_datas:
            role = Role.from_data(role_data, guild_id, strong_cache = False)
            roles[role.id] = role
    
    return roles

put_roles = entity_dictionary_putter_factory('roles', Role, force_include_internals = True)
validate_roles = entity_dictionary_validator_factory('roles', Role)

# rules_channel_id

parse_rules_channel_id = entity_id_parser_factory('rules_channel_id')
put_rules_channel_id = entity_id_optional_putter_factory('rules_channel_id')
validate_rules_channel_id = entity_id_validator_factory('rules_channel_id', Channel)

# safety_alerts_channel_id

parse_safety_alerts_channel_id = entity_id_parser_factory('safety_alerts_channel_id')
put_safety_alerts_channel_id = entity_id_optional_putter_factory('safety_alerts_channel_id')
validate_safety_alerts_channel_id = entity_id_validator_factory('safety_alerts_channel_id', Channel)


# soundboard_sounds

def parse_soundboard_sounds(data, soundboard_sounds):
    """
    Parses the guild's soundboard_sounds from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    
    soundboard_sounds : `None | dict<int, SoundboardSound>`
        The guild's soundboard sounds.
        
    Returns
    -------
    soundboard_sounds : `None | dict<int, SoundboardSound>`
        Returns `soundboard sounds` parameter.
    """
    if (soundboard_sounds is not None):
        soundboard_sounds.clear()
        soundboard_sounds = None
    
    soundboard_sound_datas = data.get('soundboard_sounds', None)
    if (soundboard_sound_datas is not None):
        for soundboard_sound_data in soundboard_sound_datas:
            soundboard_sound = SoundboardSound.from_data(soundboard_sound_data, strong_cache = False)
            if (soundboard_sound is None):
                continue
            
            if soundboard_sounds is None:
                soundboard_sounds = {}
            
            soundboard_sounds[soundboard_sound.id] = soundboard_sound
    
    return soundboard_sounds


put_soundboard_sounds = entity_dictionary_putter_factory(
    'soundboard_sounds', SoundboardSound, force_include_internals = True
)
validate_soundboard_sounds = nullable_entity_dictionary_validator_factory('soundboard_sounds', SoundboardSound)

# scheduled_events

def parse_scheduled_events(data, scheduled_events):
    """
    Parses the guild's scheduled_events from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    scheduled_events : `dict<int, ScheduledEvent>`
        The guild's scheduled_events.
        
    Returns
    -------
    scheduled_events : `dict<int, ScheduledEvent>`
        Returns `scheduled_events` parameter.
    """
    if (scheduled_events is not None):
        old_scheduled_events = [*scheduled_events.values()]
        scheduled_events = None
    
    scheduled_event_datas = data.get('guild_scheduled_events', None)
    if (scheduled_event_datas is not None):
        for scheduled_event_data in scheduled_event_datas:
            scheduled_event = ScheduledEvent.from_data(scheduled_event_data, strong_cache = False)
            
            if scheduled_events is None:
                scheduled_events = {}
            
            scheduled_events[scheduled_event.id] = scheduled_event
    
    return scheduled_events

put_scheduled_events = entity_dictionary_putter_factory(
    'guild_scheduled_events', ScheduledEvent, force_include_internals = True
)
validate_scheduled_events = nullable_entity_dictionary_validator_factory('scheduled_events', ScheduledEvent)

# stages

def parse_stages(data, stages):
    """
    Parses the guild's stages from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    stages : `None`, `dict<int, Stage>`
        The guild's stages.
        
    Returns
    -------
    stages : `None`, `dict<int, Stage>`
        Returns `stages` parameter.
    """
    if (stages is not None):
        old_stages = [*stages.values()]
        stages.clear()
    
    stage_datas = data.get('stage_instances', None)
    if (stage_datas is not None):
        for stage_data in stage_datas:
            stage = Stage.from_data(stage_data, strong_cache = False)
            
            if stages is None:
                stages = {}
            
            stages[stage.id] = stage
    
    if (stages is not None) and (not stages):
        stages = None
    
    return stages


put_stages = entity_dictionary_putter_factory(
    'stage_instances', Stage, force_include_internals = True
)
validate_stages = nullable_entity_dictionary_validator_factory('stages', Stage)

# stickers

def parse_stickers(data, stickers):
    """
    Parses the guild's stickers from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    stickers : `dict<int, Sticker>`
        The guild's stickers.
        
    Returns
    -------
    stickers : `dict<int, Sticker>`
        Returns `stickers` parameter.
    """
    if stickers:
        old_stickers = [*stickers.values()]
        stickers.clear()
    
    sticker_datas = data.get('stickers', None)
    if (sticker_datas is not None):
        for sticker_data in sticker_datas:
            sticker = Sticker.from_data(sticker_data)
            stickers[sticker.id] = sticker
    
    return stickers


put_stickers = entity_dictionary_putter_factory('stickers', Sticker, force_include_internals = True)
validate_stickers = entity_dictionary_validator_factory('stickers', Sticker)

# system_channel_flags

parse_system_channel_flags = flag_parser_factory(
    'system_channel_flags', SystemChannelFlag, default_value = SystemChannelFlag.NONE
)
put_system_channel_flags = flag_putter_factory('system_channel_flags')
validate_system_channel_flags = flag_validator_factory(
    'system_channel_flags', SystemChannelFlag, default_value = SystemChannelFlag.NONE
)

# system_channel_id

parse_system_channel_id = entity_id_parser_factory('system_channel_id')
put_system_channel_id = entity_id_optional_putter_factory('system_channel_id')
validate_system_channel_id = entity_id_validator_factory('system_channel_id', Channel)

# threads

def parse_threads(data, threads, guild_id = 0):
    """
    Parses the guild's threads from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    threads : `None | dict<int, Channel>`
        The guild's threads.
    guild_id : `int` = `0`, Optional
        The guild's identifier.
        
    Returns
    -------
    threads : `None | dict<int, Channel>`
        Returns `threads` parameter.
    """
    if threads:
        old_threads = [*threads.values()]
        threads = None
    
    thread_datas = data.get('threads', None)
    if (thread_datas is not None):
        for thread_data in thread_datas:
            thread = Channel.from_data(thread_data, None, guild_id, strong_cache = False)
            if threads is None:
                threads = {}
            
            threads[thread.id] = thread
    
    return threads


put_threads = entity_dictionary_putter_factory('threads', Channel, force_include_internals = True)
validate_threads = nullable_entity_dictionary_validator_factory('threads', Channel)

# user_count

parse_user_count = int_parser_factory('member_count', 0)
put_user_count = int_putter_factory('member_count')
validate_user_count = int_conditional_validator_factory(
    'user_count',
    0,
    lambda user_count : user_count >= 0,
    '>= 0',
)

# users

def parse_users__cache_presence(data, users, guild_id = 0):
    if users:
        old_users = [*users.values()]
        users.clear()
    
    guild_profile_datas = data.get('members', None)
    if (guild_profile_datas is not None):
        for guild_profile_data in guild_profile_datas:
            user = User.from_data(guild_profile_data['user'], guild_profile_data, guild_id, strong_cache = False)
            users[user.id] = user
    
    presence_datas = data.get('presences', None)
    if (presence_datas is not None):
        for presence_data in presence_datas:
            user_id = int(presence_data['user']['id'])
            try:
                user = users[user_id]
            except KeyError:
                pass
            else:
                user._update_presence(presence_data)
    
    return users


def parse_users__no_cache_presence(data, users, guild_id = 0):
    users.clear()
    return users


if CACHE_PRESENCE:
    parse_users = parse_users__cache_presence
else:
    parse_users = parse_users__no_cache_presence


set_docs(
    parse_users,
    """
    Parses the guild's users from the given data.
    
    > Used when presence caching is enabled.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    users : `dict<int, ClientUserBase>`
        The guild's users.
    guild_id : `int` = `0`, Optional
        The guild's identifier.
        
    Returns
    -------
    users : `dict<int, ClientUserBase>`
        Returns `users` parameter.
    """
)


def put_users(users, data, defaults, *, guild_id = 0):
    """
    Puts the given `users` into the given `data` json serializable object.
    
    Parameters
    ----------
    users : `dict`, ``WeakValueDictionary``  of (`int`, ``ClientUserBase``) items
        Resolved users.
    data : `dict<str, object>`
        Interaction resolved data.
    defaults : `bool`
        Whether default fields values should be included as well.
    guild_id : `int` = `0`, Optional (keyword only)
        The guild's identifier.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    guild_profiles_datas = []
    
    if guild_id:
        for user in users.values():
            try:
                guild_profile = user.guild_profiles[guild_id]
            except KeyError:
                continue
            
            guild_profile_data = guild_profile.to_data(defaults = defaults, include_internals = True)
            guild_profile_data['user'] = user.to_data(defaults = defaults, include_internals = True)
            guild_profiles_datas.append(guild_profile_data)
            continue
    
    data['members'] = guild_profiles_datas
    
    return data


validate_users = entity_dictionary_validator_factory('users', ClientUserBase)

# vanity_code

parse_vanity_code = nullable_string_parser_factory('vanity_url_code')
put_vanity_code = nullable_string_optional_putter_factory('vanity_url_code')
validate_vanity_code = nullable_string_validator_factory('vanity_code', 0, 1024)

# verification_level

parse_verification_level = preinstanced_parser_factory('verification_level', VerificationLevel, VerificationLevel.none)
put_verification_level = preinstanced_putter_factory('verification_level')
validate_verification_level = preinstanced_validator_factory('verification_level', VerificationLevel)

# voice_states

def parse_voice_states(data, voice_states, guild_id = 0):
    """
    Parses the guild's voice states from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    
    voice_states : `None | dict<int, VoiceState>`
        The guild's voice states.
    
    guild_id : `int` = `0`, Optional
        The guild's identifier.
        
    Returns
    -------
    voice_states : `None | dict<int, VoiceState>`
        Returns `voice_states` parameter.
    """
    if (voice_states is not None):
        voice_states.clear()
        voice_states = None
    
    voice_state_datas = data.get('voice_states', None)
    if (voice_state_datas is not None):
        for voice_state_data in voice_state_datas:
            voice_state = VoiceState.from_data(voice_state_data, guild_id, strong_cache = False)
            if (voice_state is None):
                continue
            
            if voice_states is None:
                voice_states = {}
            
            voice_states[voice_state.user_id] = voice_state
    
    return voice_states


put_voice_states = entity_dictionary_putter_factory('voice_states', VoiceState)


def validate_voice_states(field_value):
    """
    Validates the given voice states field.
    
    Parameters
    ----------
    field_value : `None`, `iterable` of ``VoiceState``, `dict` of (`int`, ``VoiceState``) items
        The field value to validate.
    
    Returns
    -------
    entity_dictionary : `None | dict<int, VoiceState>`
    
    Raises
    ------
    TypeError
        - If `field_value`'s or it's elements type is incorrect.
    """
    validated = None
    
    if field_value is not None:
        if isinstance(field_value, dict):
            iterator = iter(field_value.values())
        
        elif (getattr(field_value, '__iter__', None) is not None):
            iterator = iter(field_value)
        
        else:
            raise TypeError(
                f'`voice_states` can be `None`, `dict` of (`int`, `{VoiceState.__name__}`) items, `iterable` of '
                f'`{VoiceState.__name__}` items, got {type(field_value).__name__}; {field_value!r}.'
            )
        
        for voice_state in iterator:
            if not isinstance(voice_state, VoiceState):
                raise TypeError(
                    f'`voice_states` elements can be `{VoiceState.__name__}`, got {type(voice_state).__name__}; '
                    f'{voice_state!r}; voice_states = {field_value!r}.'
                )
            
            if validated is None:
                validated = {}
            
            validated[voice_state.user_id] = voice_state
    
    return validated

# widget_channel_id

parse_widget_channel_id = entity_id_parser_factory('widget_channel_id')
put_widget_channel_id = entity_id_optional_putter_factory('widget_channel_id')
validate_widget_channel_id = entity_id_validator_factory('widget_channel_id', Channel)

# widget_enabled

parse_widget_enabled = bool_parser_factory('widget_enabled', False)
put_widget_enabled = bool_optional_putter_factory('widget_enabled', False)
validate_widget_enabled = bool_validator_factory('widget_enabled', False)


# ---- extra ----

def validate_channels_and_channel_datas(channels):
    """
    Validates the given `channels` value. This function is used when creating a new guild.
    
    Parameters
    ----------
    channels : `None`, `list` of (``Channel``, `dict`)
    
    Returns
    -------
    validated : `None`, `list` of (``Channel``, `dict`)
    
    Raises
    ------
    TypeError
        - If a `channel`'s type is incorrect.
    ValueError
        - If a `channel`'s channels is incorrect.
    """
    validated = None
    
    if channels is None:
        pass
    
    elif isinstance(channels, list):
        for element in channels:
            if not isinstance(element, (Channel, dict)):
                raise TypeError(
                    f'`channel` elements can be `{Channel.__name__}`, `dict`. '
                    f'Got {type(element).__name__}; {element!r}; channels = {channels!r}'
                )
        
            if validated is None:
                validated = []
            
            validated.append(element)
    
    else:
        raise TypeError(
            f'`channels` can be `None` or `dict` of `{Channel.__name__}`, `dict` elements. '
            f'Got {channels.__class__.__name__}; {channels!r}.'
        )
    
    return validated


def validate_roles_and_role_datas(roles):
    """
    Validates the given `roles` value. This function is used when creating a new guild.
    
    Parameters
    ----------
    roles : `None`, `list` of (``Role``, `dict`)
    
    Returns
    -------
    validated : `None`, `list` of (``Role``, `dict`)
    
    Raises
    ------
    TypeError
        - If a `role`'s type is incorrect.
    ValueError
        - If a `role`'s roles is incorrect.
    """
    validated = None
    
    if roles is None:
        pass
    
    elif isinstance(roles, list):
        for element in roles:
            if not isinstance(element, (Role, dict)):
                raise TypeError(
                    f'`role` elements can be `{Role.__name__}`, `dict`. '
                    f'Got {type(element).__name__}; {element!r}; roles = {roles!r}'
                )
        
            if validated is None:
                validated = []
            
            validated.append(element)
    
    else:
        raise TypeError(
            f'`roles` can be `None` or `dict` of `{Role.__name__}`, `dict` elements. '
            f'Got {roles.__class__.__name__}; {roles!r}.'
        )
    
    return validated


def put_channels_and_channel_datas(channels, data, defaults):
    """
    Puts the given channels or their data into the given `data` json serializable object.
    
    Parameters
    ----------
    channels : `None`, `list` of (``Channel``, `dict`)
        Channels or channel datas.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    channel_datas = []
    
    if (channels is not None):
        for channel in channels:
            if isinstance(channel, Channel):
                channel_data = channel.to_data(defaults = defaults, include_internals = True)
            else:
                channel_data = channel
            channel_datas.append(channel_data)
    
    data['channels'] = channel_datas
    return data


def put_roles_and_role_datas(roles, data, defaults):
    """
    Puts the given roles or their data into the given `data` json serializable object.
    
    Parameters
    ----------
    roles : `None`, `list` of (``Role``, `dict`)
        Roles or role datas.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    role_datas = []
    
    if (roles is not None):
        for role in roles:
            if isinstance(role, Role):
                role_data = role.to_data(defaults = defaults, include_internals = True)
            else:
                role_data = role
            role_datas.append(role_data)
    
    data['roles'] = role_datas
    return data



__all__ = ()

from scarletio import Compound

from ...application import Application
from ...http import DiscordApiClient

from ...emoji import Emoji
from ...emoji.emoji.utils import EMOJI_APPLICATION_FIELD_CONVERTERS
from ...http import DiscordApiClient
from ...payload_building import build_create_payload, build_edit_payload
from ...utils import image_to_base64

from ..request_helpers import get_emoji_and_id, get_emoji_id

from .interaction import _assert__application_id


class ClientCompoundEmojiApplicationEndpoints(Compound):
    
    api : DiscordApiClient
    application: Application
    
    async def emoji_get_application(self, emoji, *, force_update = False):
        """
        Requests the emoji by it's id at the given application.
        If the client's logging in is finished, then it should have all of its emojis loaded already.
        
        This method is a coroutine.
        
        Parameters
        ----------
        emoji : ``Emoji``, `(int, int)`
            The emoji, or 2 snowflake representing it.
        
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the emoji should be requested even if it supposed to be up to date.
        
        Returns
        -------
        emoji : ``Emoji``
        
        Raises
        ------
        TypeError
            - If `emoji` is given as incorrect type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        emoji, emoji_id = get_emoji_and_id(emoji)
        
        if (not force_update) and (emoji is not None) and (self.application._has_cache_emoji_by_id(emoji_id)):
            return emoji
        
        emoji_data = await self.api.emoji_get_application(application_id, emoji_id)
        
        if (emoji is None):
            emoji = Emoji.from_data(emoji_data, 0)
        else:
            emoji._set_attributes(emoji_data, 0)
        
        self.application._add_cache_emoji(emoji)
        return emoji
    
    
    async def emoji_get_all_application(self):
        """
        Requests the given application's emojis.
        
        This method is a coroutine.
        
        Returns
        -------
        emojis : `list` of ``Emoji``
            The application's emojis.
        
        Raises
        ------
        TypeError
            - If `application`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        data = await self.api.emoji_get_all_application(application_id)
        emoji_datas = data.get('items', None)
        
        emojis = []
        if (emoji_datas is not None):
            for emoji_data in emoji_datas:
                emoji = Emoji.from_data(emoji_data, 0)
                self.application._add_cache_emoji(emoji)
                emojis.append(emoji)
        
        return emojis
    
    
    async def emoji_create_application(self, image, emoji_template = None, **keyword_parameters):
        """
        Creates an emoji at the client's application.
        
        This method is a coroutine.
        
        Parameters
        ----------
        image : `bytes-like`
            The emoji's icon. Up to 256 kB.
        
        emoji_template : ``None | Emoji`` = `None`, Optional
            Channel entity to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the emoji with.
        
        Other attributes
        ----------------
        name : `str`, Optional (Keyword only)
            The emoji's name. It's length can be between `2` and `32`.
        
        Returns
        -------
        emoji : ``Emoji``
            The created emoji.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        Only some characters can be in the emoji's name, so every other character is filtered out.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        data = build_create_payload(emoji_template, EMOJI_APPLICATION_FIELD_CONVERTERS, keyword_parameters)
        data['image'] = image_to_base64(image)
        emoji_data = await self.api.emoji_create_application(application_id, data)
        
        emoji = Emoji.from_data(emoji_data, 0)
        self.application._add_cache_emoji(emoji)
        return emoji
    
    
    async def emoji_edit_application(self, emoji, emoji_template = None, **keyword_parameters):
        """
        Edits the given (application) emoji.
        
        This method is a coroutine.
        
        Parameters
        ----------
        emoji : ``Emoji``, `(int, int)`
            The emoji to edit.
        
        emoji_template : ``None | Emoji`` = `None`, Optional
            A emoji to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other attributes
        ----------------
        name : `str`, Optional (Keyword only)
            The emoji's name. It's length can be between `2` and `32`.
        
        Returns
        -------
        emoji : ``Emoji``
            The created emoji.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        emoji, emoji_id = get_emoji_and_id(emoji)
        
        data = build_edit_payload(emoji, emoji_template, EMOJI_APPLICATION_FIELD_CONVERTERS, keyword_parameters)
        
        if not data:
            return
        
        emoji_data = await self.api.emoji_edit_application(application_id, emoji_id, data)
        
        if emoji is None:
            emoji = Emoji.from_data(emoji_data, 0)
        else:
            emoji._set_attributes(emoji_data, 0)
        
        self.application._add_cache_emoji(emoji)
    
    
    async def emoji_delete_application(self, emoji):
        """
        Creates an emoji at the client's application.
        
        This method is a coroutine.
        
        Parameters
        ----------
        emoji : ``Emoji``, `int`
            The emoji to delete.
        
        Raises
        ------
        TypeError
            - If emoji is given as incorrect type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        emoji_id = get_emoji_id(emoji)
        await self.api.emoji_delete_application(application_id, emoji_id)
        
        self.application._delete_cache_emoji_by_id(emoji_id)



__all__ = ('parse_oauth2_redirect_url',)

import re


OAUTH2_REQUEST_URL_RP = re.compile('(https?://.+?)\\?code=([a-zA-Z0-9]{30})')


def parse_oauth2_redirect_url(url):
    """
    Parses the `redirect_url` and the `code` out from a whole `url`, what an user was redirected to after oauth2
    authorization.
    
    If the parsing was successful, then returns a `tuple` of `redirect_url` and `code`. If it fails, returns `None`.
    
    Parameters
    ----------
    url : `str`
        A whole to url to parse from

    Returns
    -------
    result : `None`, `tuple` (`str`, `str`)
    """
    result = OAUTH2_REQUEST_URL_RP.fullmatch(url)
    if result is None:
        return None
    
    return result.groups()



__all__ = ()

from scarletio import copy_docs

from ..converter_constants import ANNOTATION_TYPE_TO_STR_ANNOTATION

from .base import ParameterConverterBase


class ParameterConverterInternal(ParameterConverterBase):
    """
    Internal parameter converter.
    
    Attributes
    ----------
    converter : `CoroutineFunctionType`
        The converter to use to convert a value to it's desired type.
    
    parameter_name : `str`
        The parameter's name.
    
    type : `int`
        Internal identifier of the converter.
    """
    __slots__ = ('converter', 'type')
    
    def __new__(cls, parameter_name, converter_type, converter):
        """
        Creates a new ``ParameterConverterInternal`` with the given parameters.
        
        Parameters
        ----------
        parameter_name : `str`
            The parameter's name.
        
        converter_type : `int`
            Internal identifier of the converter.
        
        converter : `CoroutineFunctionType`
            The converter to use to convert a value to it's desired type.
        """
        self = object.__new__(cls)
        self.parameter_name = parameter_name
        self.type = converter_type
        self.converter = converter
        return self
    
    
    @copy_docs(ParameterConverterBase.__call__)
    async def __call__(self, client, interaction_event, value):
        return await self.converter(client, interaction_event)
    
    
    @copy_docs(ParameterConverterBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        # type
        repr_parts.append(', type = ')
        repr_parts.append(ANNOTATION_TYPE_TO_STR_ANNOTATION[self.type])



__all__ = ()

from scarletio import Compound

from ...channel import Channel
from ...http import DiscordApiClient
from ...payload_building import build_create_payload, build_edit_payload
from ...stage import Stage
from ...stage.stage.utils import STAGE_CREATE_FIELD_CONVERTERS, STAGE_EDIT_FIELD_CONVERTERS

from ..request_helpers import get_channel_id, get_stage_and_channel_id, get_stage_channel_id


class ClientCompoundStageEndpoints(Compound):
    
    api : DiscordApiClient
    
    
    async def stage_create(self, channel, stage_template = None, *, reason = None, **keyword_parameters):
        """
        Edits the given stage channel.
        
        Will trigger a `stage_create` event.
        
        > The endpoint has a long rate limit, so please use ``.stage_edit`` to edit just the stage's topic.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to create the stage at.
        
        stage_template : `None`, ``Stage`` = `None`, Optional
            Stage entity to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the role with.
        
        Other Parameters
        ----------------
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The new privacy level of the stage. Defaults to guild only.
        
        scheduled_event : `int`, ``ScheduledEvent``, Optional (Keyword only)
            Alternative for `scheduled_event_id`.
        
        scheduled_event_id : `int`, ``ScheduledEvent``, Optional (Keyword only)
            The scheduled event's identifier that started the stage.
        
        send_start_notification : `bool`, Optional (Keyword only)
            Whether @everyone should be notified when the stage is started.
            
            > You must have `mention everyone` permission.
        
        topic : `None`, `str`, Optional (Keyword only)
            The topic of the stage.
        
        Returns
        -------
        stage : ``Stage``
            The created stage instance.
        
        Raises
        ------
        TypeError
            - If `channel` was not given as ``Channel`` neither as `int`.
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        # channel_id
        channel_id = get_channel_id(channel, Channel.is_guild_stage)
        
        data = build_create_payload(stage_template, STAGE_CREATE_FIELD_CONVERTERS, keyword_parameters)
        data['channel_id'] = channel_id
        
        stage_data = await self.api.stage_create(data, reason)
        return Stage.from_data(stage_data)
    
    
    async def stage_edit(self, stage, stage_template = None, *, reason = None, **keyword_parameters):
        """
        Edits the given stage channel.
        
        Will trigger a `stage_edit` event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        stage : ``Stage``, ``Channel``, `int`
            The stage to edit. Can be given as it's channel's identifier.
        
        stage_template : `None`, ``Stage`` = `None`, Optional
            Stage entity to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the role with.
        
        Other Parameters
        ----------------
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The new privacy level of the stage.
        
        topic : `None`, `str`, Optional (Keyword only)
            The new topic of the stage.
        
        Raises
        ------
        TypeError
            - If `channel` was not given as ``Channel`` neither as `int`.
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        stage, channel_id = get_stage_and_channel_id(stage)
        
        data = build_edit_payload(stage, stage_template, STAGE_EDIT_FIELD_CONVERTERS, keyword_parameters)
        
        if data:
            await self.api.stage_edit(channel_id, data, reason)
            # We receive data, but ignore it, so we can dispatch it.
    
    
    async def stage_delete(self, stage, *, reason = None):
        """
        Deletes the given stage channel.
        
        Will trigger a `stage_delete` event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        stage : ``Stage``, ``Channel``, `int`
            The stage to delete. Can be given as it's channel's identifier.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            If `stage` was not given as ``Stage``, ``Channel`` neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_stage_channel_id(stage)
        
        await self.api.stage_delete(channel_id, reason)
        # We receive no data.
    
    
    async def stage_get(self, channel):
        """
        Gets the stage of the given stage channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The stage's channel's identifier.
        
        Raises
        ------
        TypeError
            If `channel` was not given as ``Channel`` neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_guild_stage)
        
        data = await self.api.stage_get(channel_id)
        return Stage.from_data(data)



__all__ = ()

from ...color import Color


def value_deserializer_id(value):
    """
    Shared `get_id` conversion.
    
    Parameters
    ----------
    value : `None | str`
        Raw value to convert.
    
    Returns
    -------
    value : `int`
    """
    if value is None:
        value = 0
    else:
        value = int(value)
    
    return value


def value_serializer_id(value):
    """
    Shared `put_id` conversion.
    
    Parameters
    ----------
    value : `int`
        Processed value to convert.
    
    Returns
    -------
    value : `None | str`
    """
    if value:
        value = str(value)
    else:
        value = None
    
    return value


def value_deserializer_ids(value):
    """
    Shared `get_ids` conversion.
    
    Parameters
    ----------
    value : `None | list<str>`
        Raw value to convert.
    
    Returns
    -------
    value : `None | tuple<int>`
    """
    if value is None:
        pass
    elif not value:
        value = None
    else:
        value = (*sorted(int(entity_id) for entity_id in value),)
    
    return value


def value_serializer_ids(value):
    """
    Shared `put_ids` conversion.
    
    Parameters
    ----------
    value : `None | tuple<int>`
        Processed value to convert.
    
    Returns
    -------
    value : `list<str>`
    """
    if value is None:
        value = []
    else:
        value = [str(entity_id) for entity_id in value]
    
    return value


def value_deserializer_name(value):
    """
    Shared `get_name` conversion.
    
    Parameters
    ----------
    value : `None | str`
        Raw value to convert.
    
    Returns
    -------
    value : `str`
    """
    if value is None:
        value = ''
    
    return value


def value_serializer_name(value):
    """
    Shared `put_name` conversion.
    
    Parameters
    ----------
    value : `str`
        Processed value to convert.
    
    Returns
    -------
    value : `str`
    """
    return value


def value_deserializer_string_array(value):
    """
    Shared `get_string_array` conversion.
    
    Parameters
    ----------
    value : `None | list<str>`
        Raw value to convert.
    
    Returns
    -------
    value : `None | tuple<str>`
    """
    if value is None:
        pass
    elif (not value):
        value = None
    else:
        value = (*sorted(value),)
    return value


def value_serializer_string_array(value):
    """
    Shared `put_string_array` conversion.
    
    Parameters
    ----------
    value : `None | tuple<str>`
        Processed value to convert.
    
    Returns
    -------
    value : `list<str>`
    """
    if value is None:
        value = []
    else:
        value = [*value]
    return value


def value_deserializer_description(value):
    """
    Shared `get_description` conversion.
    
    Parameters
    ----------
    value : `None | str`
        Raw value to convert.
    
    Returns
    -------
    value : `None | str`
    """
    if value is None:
        pass
    elif (not value):
        value = None
    return value


def value_serializer_description(value):
    """
    Shared `put_description` conversion.
    
    Parameters
    ----------
    value : `None | str`
        Processed value to convert.
    
    Returns
    -------
    value : `str`
    """
    if value is None:
        value = ''
    
    return value


def value_deserializer_html_color(value):
    """
    Shared `get_html_color` conversion.
    
    Parameters
    ----------
    value : `None | str`
        Raw value to convert.
    
    Returns
    -------
    value : `None | Color`
    """
    if value is not None:
        value = Color.from_html(value)
    
    return value


def value_serializer_html_color(value):
    """
    Shared `put_html_color` conversion.
    
    Parameters
    ----------
    value : `None | Color`
        Processed value to convert.
    
    Returns
    -------
    value : `None | str`
    """
    if (value is not None):
        value = value.as_html
    
    return value



from .constants import *
from .fields import *
from .preinstanced import *
from .shared_client_theme import *


__all__ = (
    *constants.__all__,
    *fields.__all__,
    *preinstanced.__all__,
    *shared_client_theme.__all__,
)



__all__ = ('ScheduledEventSubscribeEvent',)

from scarletio import copy_docs

from ...bases import EventBase
from ...core import GUILDS, SCHEDULED_EVENTS
from ...user import create_partial_user_from_id

from .fields import (
    parse_guild_id, parse_scheduled_event_id, parse_user_id, put_guild_id, put_scheduled_event_id,
    put_user_id, validate_guild_id, validate_scheduled_event_id, validate_user_id
)


class ScheduledEventSubscribeEvent(EventBase):
    """
    Represents a `GUILD_SCHEDULED_EVENT_USER_ADD` event.
    
    Attributes
    ----------
    guild_id : `int`
        The guild's identifier where the event is for.
    
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    user_id : `int`
        The identifier of the user, who subscribed to the event.
    """
    __slots__ = ('guild_id', 'scheduled_event_id', 'user_id', )
    
    
    def __new__(cls, *, guild_id = ..., scheduled_event_id = ..., user_id = ...):
        """
        Creates a new scheduled event (un)subscribe event.
        
        Parameters
        ----------
        guild_id : ˙`None | int | Guild``, Optional (Keyword only)
            The guild or its identifier where the event will be.
        
        scheduled_event_id : ``None | int | ScheduledEvent``, Optional (Keyword only)
            The scheduled event or its identifier.
        
        user_id : ``None | int | ClientUserBase``, Optional (Keyword only)
            The user or their their identifier who subscribed to the event.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # scheduled_event_id
        if scheduled_event_id is ...:
            scheduled_event_id = 0
        else:
            scheduled_event_id = validate_scheduled_event_id(scheduled_event_id)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.scheduled_event_id = scheduled_event_id
        self.user_id = user_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new scheduled event (un)subscribe event from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Scheduled event subscribe event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.guild_id = parse_guild_id(data)
        self.scheduled_event_id = parse_scheduled_event_id(data)
        self.user_id = parse_user_id(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the scheduled event subscribe event.
        
        Parameters
        ----------
        defaults : `bool`
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_guild_id(self.guild_id, data, defaults)
        put_scheduled_event_id(self.scheduled_event_id, data, defaults)
        put_user_id(self.user_id, data, defaults)
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' guild_id = ',
            repr(self.guild_id),
            ', scheduled_event_id = ',
            repr(self.scheduled_event_id),
            ', user_id = ',
            repr(self.user_id),
            '>'
        ]
        
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.scheduled_event_id
        yield self.user_id
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.scheduled_event_id != other.scheduled_event_id:
            return False
        
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        return self.guild_id ^ self.scheduled_event_id ^ self.user_id
    

    def copy(self):
        """
        Copies the scheduled event (un)subscribe event.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.guild_id = self.guild_id
        new.scheduled_event_id = self.scheduled_event_id
        new.user_id = self.user_id
        return new
    
    
    def copy_with(self, guild_id = ..., scheduled_event_id = ..., user_id = ...):
        """
        Copies the scheduled event (un)subscribe event with the given fields.
        
        Parameters
        ----------
        guild_id : ˙`None | int | Guild``, Optional (Keyword only)
            The guild or its identifier where the event will be.
        
        scheduled_event_id : ``None | int | ScheduledEvent``, Optional (Keyword only)
            The scheduled event or its identifier.
        
        user_id : ``None | int | ClientUserBase``, Optional (Keyword only)
            The user or their their identifier who subscribed to the event.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # scheduled_event_id
        if scheduled_event_id is ...:
            scheduled_event_id = self.scheduled_event_id
        else:
            scheduled_event_id = validate_scheduled_event_id(scheduled_event_id)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        new = object.__new__(type(self))
        new.guild_id = guild_id
        new.scheduled_event_id = scheduled_event_id
        new.user_id = user_id
        return new
    
    
    @property
    def scheduled_event(self):
        """
        Returns the scheduled event.
        
        Returns
        -------
        scheduled_event : ``None | ScheduledEvent``
        """
        return SCHEDULED_EVENTS.get(self.scheduled_event_id, None)
    
    
    @property
    def guild(self):
        """
        Returns the scheduled event subscribe event's guild.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def user(self):
        """
        Returns the user who subscribed to the event.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        return create_partial_user_from_id(self.user_id)



__all__ = ()

from ...field_parsers import int_parser_factory, preinstanced_array_parser_factory
from ...field_putters import int_putter_factory, preinstanced_array_putter_factory
from ...field_validators import int_conditional_validator_factory, preinstanced_array_validator_factory
from ...guild import GuildFeature


# additional_emoji_slots

parse_additional_emoji_slots = int_parser_factory('additional_emoji_slots', 0)
put_additional_emoji_slots = int_putter_factory('additional_emoji_slots')
validate_additional_emoji_slots = int_conditional_validator_factory(
    'additional_emoji_slots',
    0,
    (lambda additional_emoji_slots : additional_emoji_slots >= 0),
    '>= 0',
)

# additional_soundboard_sound_slots

parse_additional_soundboard_sound_slots = int_parser_factory('additional_sound_slots', 0)
put_additional_soundboard_sound_slots = int_putter_factory('additional_sound_slots')
validate_additional_soundboard_sound_slots = int_conditional_validator_factory(
    'additional_soundboard_sound_slots',
    0,
    (lambda additional_soundboard_sound_slots : additional_soundboard_sound_slots >= 0),
    '>= 0',
)

# additional_sticker_slots

parse_additional_sticker_slots = int_parser_factory('additional_sticker_slots', 0)
put_additional_sticker_slots = int_putter_factory('additional_sticker_slots')
validate_additional_sticker_slots = int_conditional_validator_factory(
    'additional_sticker_slots',
    0,
    (lambda additional_sticker_slots : additional_sticker_slots >= 0),
    '>= 0',
)

# features

parse_features = preinstanced_array_parser_factory('features', GuildFeature)
put_features = preinstanced_array_putter_factory('features')
validate_features = preinstanced_array_validator_factory('features', GuildFeature)



__all__ = ('ApplicationCommandOptionMetadataNested',)

from scarletio import copy_docs

from .base import ApplicationCommandOptionMetadataBase
from .fields import parse_options, put_options, validate_options


class ApplicationCommandOptionMetadataNested(ApplicationCommandOptionMetadataBase):
    """
    Base type for nested application command option metadatas.
    
    Attributes
    ----------
    options : `None`, `tuple` of ``ApplicationCommandOption``
        Contains the option's parameter or its sub-commands (or groups).
    """
    __slots__ = ('options',)
    
    
    def __new__(cls, *, options = ...):
        """
        Creates a new nested application command option metadata with the given parameters.
        
        Parameters
        ----------
        options : `None`, `iterable` of ``ApplicationCommandOption``, Optional (Keyword only)
            Contains the option's parameter or its sub-commands (or groups).
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # options
        if options is ...:
            options = None
        else:
            options = validate_options(options)
        
        # Construct
        new = object.__new__(cls)
        new.options = options
        return new
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            options = keyword_parameters.pop('options', ...),
        )
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.options = parse_options(data)
        return self
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_options(self.options, data, defaults)
        return data
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase._add_type_specific_repr_fields)
    def _add_type_specific_repr_fields(self, repr_parts):
        repr_parts.append(' options = ')
        repr_parts.append(repr(self.options))
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if self.options != other.options:
            return False
        
        return True
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options) << 10
            
            for option in options:
                hash_value ^= hash(options)
        
        return hash_value
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        options = self.options
        if (options is not None):
            options = (*(option.copy() for option in options),)
        new.options = options
        return new
    
    
    def copy_with(self, *, options = ...):
        """
        Copies the nested application command option metadata with the given fields.
        
        Parameters
        ----------
        options : `None`, `iterable` of ``ApplicationCommandOption``, Optional (Keyword only)
            Contains the option's parameter or its sub-commands (or groups).
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # options
        if options is ...:
            options = self.options
            if (options is not None):
                options = (*(option.copy() for option in options),)
        else:
            options = validate_options(options)
        
        # Construct
        new = object.__new__(type(self))
        new.options = options
        return new
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            options = keyword_parameters.pop('options', ...),
        )



__all__ = ('CONVERSION_ABORT',)

from ....discord.builder.constants import CONVERSION_KIND_FIELD
from ....discord.builder.conversion import Conversion


class CONVERSION_ABORT(Conversion):
    # Generic
    
    name = 'abort'
    name_aliases = None
    expected_types_messages = '`None`, `bool`'
    kind = CONVERSION_KIND_FIELD
    output_conversion = None
    
    # Setting
    
    set_merger = None
    set_identifier = None
    set_type = None
    set_type_processor = None
    set_listing_identifier = None
    
    
    def set_validator(value):
        # None
        if value is None:
            yield False
            return
        
        # bool
        if isinstance(value, bool):
            yield value
            return
        
        # No other cases
        return
    
    
    # Reading
    
    get_default = False
    get_processor = None
    
    
    # Serialization
    
    serializer_key = None
    serializer_optional = None
    serializer_required = None
    
    # Sorting
    
    sort_priority = -1



__all__ = ('OrientationLockState', 'PlatformType',)

from ...bases import Preinstance as P, PreinstancedBase


class OrientationLockState(PreinstancedBase, value_type = int):
    """
    Represents a embedded activity's orientation lock state.
    
    Attributes
    ----------
    name : `str`
        The name of the orientation lock state.
    
    value : `int`
        The identifier value the orientation lock state.
    
    Type Attributes
    ---------------
    Every predefined orientation lock state can be accessed as type attribute as well:
    
    +-----------------------+-----------+-------+
    | Type attribute name   | Name      | Value |
    +=======================+===========+=======+
    | none                  | none      | 0     |
    +-----------------------+-----------+-------+
    | unlocked              | unlocked  | 1     |
    +-----------------------+-----------+-------+
    | portrait              | portrait  | 2     |
    +-----------------------+-----------+-------+
    | landscape             | landscape | 3     |
    +-----------------------+-----------+-------+
    """
    __slots__ = ()
    
    none = P(0, 'none')
    unlocked = P(1, 'unlocked')
    portrait = P(2, 'portrait')
    landscape = P(3, 'landscape')


class PlatformType(PreinstancedBase, value_type = str):
    """
    Represents a supported platform by an embedded activity.
    
    Attributes
    ----------
    name : `str`
        The name of the platform type.
    
    value : `str`
        The Discord side identifier value of the platform.
    
    Type Attributes
    ---------------
    Every predefined platform type can also be accessed as type attribute:
    
    +-----------------------+---------------+-----------+
    | Type attribute name   | name          | value     |
    +=======================+===============+===========+
    | android               | android       | android   |
    +-----------------------+---------------+-----------+
    | ios                   | ios           | ios       |
    +-----------------------+---------------+-----------+
    | web                   | web           | web       |
    +-----------------------+---------------+-----------+
    """
    __slots__ = ()

    # predefined
    android = P('android', 'android')
    ios = P('ios', 'ios')
    web = P('web', 'web')


PlatformType.INSTANCES[''] = PlatformType.web



__all__ = ()

from ...field_parsers import preinstanced_parser_factory, nullable_date_time_parser_factory
from ...field_putters import preinstanced_putter_factory, nullable_date_time_putter_factory
from ...field_validators import preinstanced_validator_factory, nullable_date_time_validator_factory

from .preinstanced import LabelType, ReleasePhase


# label_type

parse_label_type = preinstanced_parser_factory('label_type', LabelType, LabelType.none)
put_label_type = preinstanced_putter_factory('label_type')
validate_label_type = preinstanced_validator_factory('label_type', LabelType)

# labelled_until

parse_labelled_until = nullable_date_time_parser_factory('label_until')
put_labelled_until = nullable_date_time_putter_factory('label_until')
validate_labelled_until = nullable_date_time_validator_factory('label_until')

# release_phase

parse_release_phase = preinstanced_parser_factory('release_phase', ReleasePhase, ReleasePhase.global_launch)
put_release_phase = preinstanced_putter_factory('release_phase')
validate_release_phase = preinstanced_validator_factory('release_phase', ReleasePhase)



__all__ = ()

from ...emoji import Emoji, create_partial_emoji_from_inline_data, put_partial_emoji_inline_data_into
from ...field_parsers import force_string_parser_factory
from ...field_putters import force_string_putter_factory
from ...field_validators import force_string_validator_factory, nullable_entity_validator_factory

from .constants import TITLE_LENGTH_MAX, TITLE_LENGTH_MIN

# emoji

parse_emoji = create_partial_emoji_from_inline_data


def put_emoji(emoji, data, defaults):
    """
    Puts the emoji into the given `data` json serializable object.
    
    Parameters
    ----------
    emoji : ``None | Emoji``
        The emoji.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (emoji is not None):
        put_partial_emoji_inline_data_into(emoji, data)
    
    return data


validate_emoji = nullable_entity_validator_factory('emoji', Emoji)


# title

parse_title = force_string_parser_factory('label')
put_title = force_string_putter_factory('label')
validate_title = force_string_validator_factory('title', TITLE_LENGTH_MIN, TITLE_LENGTH_MAX)



__all__ = ()

from time import time as time_now

from scarletio import Compound, Future, Theory

from ....env import CACHE_PRESENCE

from ...activity import ACTIVITY_UNKNOWN, Activity, ActivityType
from ...bases import maybe_snowflake_pair
from ...channel import Channel
from ...core import KOKORO
from ...events.event_handler_manager import EventHandlerManager
from ...events.handling_helpers import WaitForHandler
from ...gateway.client_base import DiscordGatewayClientBase
from ...gateway.constants import (
    GATEWAY_OPERATION_CLIENT_PRESENCE, GATEWAY_OPERATION_CLIENT_REQUEST_GUILD_USERS,
    GATEWAY_OPERATION_CLIENT_REQUEST_SOUNDBOARD_SOUNDS
)
from ...preconverters import preconvert_preinstanced_type
from ...user import Status
from ...voice import VoiceClient

from ..functionality_helpers import MassUserChunker, SingleUserChunker
from ..request_helpers import get_guild_and_id, get_guild_id


def _assert__edit_presence__afk(afk):
    """
    Asserts the `afk` parameter of ``Client.edit_presence`` method.
    
    Parameters
    ----------
    afk : `bool`
        Whether the client is afk or not (?).
    
    Raises
    ------
    AssertionError
        - If `afk` is not `bool`.
    """
    if not isinstance(afk, bool):
        raise AssertionError(
            f'`afk` can be `bool`, got {type(afk).__name__}; {afk!r}.'
        )
    
    return True


def _assert__request_users__limit(limit):
    """
    Asserts the `limit` parameter of ``Client.request_users`` method.
    
    Parameters
    ----------
    limit : `int`
        The amount of users to be received. Limited to `100`.
    
    Raises
    ------
    AssertionError
        - If `limit` is not `int`.
        - If `limit` is out of the expected range [1:100].
    """
    if not isinstance(limit, int):
        raise AssertionError(
            f'`limit` can be `int`, got {type(limit).__name__}; {limit!r}.'
        )
    
    if limit < 1 or limit > 100:
        raise AssertionError(
            f'`limit` is out of the expected range [1:100], got {limit!r}.'
        )
    
    return True


def _assert__request_users__name(name):
    """
    Asserts the `name` parameter of ``Client.request_users`` method.
    
    Parameters
    ----------
    name : `str`
        The received user's name or nick should start with this string.
    
    Raises
    ------
    AssertionError
            - If `name` is not `str`.
            - If `name` length is out of the expected range [1:32].
    """
    if not isinstance(name, str):
        raise AssertionError(
            f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
        )
    
    name_length = len(name)
    if name_length < 1 or name_length > 32:
        raise AssertionError(
            f'`name` length can be in range [1:32], got {name_length!r}; {name!r}.'
        )
    
    return True


class ClientCompoundClientGateway(Compound):
    
    events : EventHandlerManager
    gateway : DiscordGatewayClientBase
    bot : bool
    voice_clients : dict
    
    @Theory
    def gateway_for(self, guild_id): ...
    
    
    async def edit_presence(self, *, activity = ..., status = ..., afk = False):
        """
        Changes the client's presence (status and activity). If a parameter is not defined, it will not be changed.
        
        This method is a coroutine.
        
        Parameters
        ----------
        activity : ``Activity``, Optional (Keyword only)
            The new activity of the Client.
        status : `str`, ``Status``, Optional (Keyword only)
            The new status of the client.
        afk : `bool` = `False`, Optional (Keyword only)
            Whether the client is afk or not (?).
        
        Raises
        ------
        TypeError:
            - If the status is not `str`, ``Status``.
            - If activity is not ``Activity`` or has type ``ActivityType.custom``.
        ValueError:
            - If the status `str`, but not any of the predefined ones.
        """
        if status is ...:
            status = self._status
        else:
            status = preconvert_preinstanced_type(status, 'status', Status)
            self._status = status
        
        status = status.value
        
        if activity is ...:
            activity = self._activity
        elif activity is None:
            self._activity = ACTIVITY_UNKNOWN
        elif isinstance(activity, Activity):
            self._activity = activity
        else:
            raise TypeError(
                f'`activity` can be `{Activity.__name__}`, got: '
                f'{type(activity).__name__}; {activity!r}.'
            )
        
        if activity is None:
            pass
        elif activity is ACTIVITY_UNKNOWN:
            activity = None
        else:
            activity = activity.to_data(user = not self.bot)
        
        if status == 'idle':
            since = int(time_now() * 1000.0)
        else:
            since = 0
        
        assert _assert__edit_presence__afk(afk)
        
        
        data = {
            'op': GATEWAY_OPERATION_CLIENT_PRESENCE,
            'd': {
                'game': activity,
                'since': since,
                'status': status,
                'afk': afk,
            },
        }
        
        await self.gateway.send_as_json(data)
    
    
    async def join_voice(self, channel):
        """
        Joins a voice client to the channel. If there is an already existing voice client at the respective guild,
        moves it.
        
        If not every library is installed, raises `RuntimeError`, or if the voice client fails to connect raises
        `TimeoutError`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `tuple` (`int`, `int`)
            The channel to join to.
        
        Returns
        -------
        voice_client : ``VoiceClient``
        
        Raises
        ------
        RuntimeError
            If not every library is installed to join voice.
        TimeoutError
            If voice client fails to connect the given channel.
        TypeError
            If `channel` was not given neither as ``Channel`` nor as `tuple` (`int`, `int`).
        """
        while True:
            if isinstance(channel, Channel):
                if channel.is_in_group_guild_connectable() or channel.partial:
                    guild_id = channel.guild_id
                    channel_id = channel.id
                    break
            
            else:
                snowflake_pair = maybe_snowflake_pair(channel)
                if (snowflake_pair is not None):
                    guild_id, channel_id = snowflake_pair
                    break
                
            raise TypeError(
                f'`channel` can be a connectable channel, `tuple` (`int`, `int`)'
                f', got {channel.__class__.__name__}; {channel!r}.'
            )
        
        
        voice_client = self.voice_clients.get(guild_id, None)
        if voice_client is None:
            voice_client = VoiceClient(self, guild_id, channel_id)
            try:
                waiter = voice_client.start()
                
                waiter.apply_timeout(60.0)
                success = await waiter
                
                if success:
                    success = await voice_client.wait_connected(60.0)
                
                if not success:
                    raise TimeoutError('Voice client failed to connect.')
            except GeneratorExit:
                raise
            
            except:
                await voice_client.disconnect()
                raise
        
        else:
            if voice_client.channel_id != channel_id:
                gateway = self.gateway_for(guild_id)
                await gateway.change_voice_state(guild_id, channel_id)
        
        return voice_client
    
    
    async def request_soundboard_sounds(self, guilds, *, force_request = False):
        """
        Requests the soundboard sounds of the given guilds.
        
        This function is a coroutine.
        
        Parameters
        ----------
        guilds : ``iterable<Guild> | iterable<int>``
            The guilds or their identifiers to request the sounds of.
        force_request : `bool` = `False`
            Whether the sounds should be requested even if they are already cached.
        
        Returns
        -------
        sounds : `None`, `list` of ``SoundboardSoundsEvent``
            Might return `None` if there are no guilds to request sounds from.
        """
        guild_ids = set()
        
        if force_request:
            for guild in guilds:
                guild_id = get_guild_id(guild)
                guild_ids.add(guild_id)
        
        else:
            for guild in guilds:
                guild, guild_id = get_guild_and_id(guild)
                if (guild is not None) and guild.soundboard_sounds_cached:
                    continue
                
                guild_ids.add(guild_id)
        
        if not guild_ids:
            return None
        
        by_gateway = {}
        
        for guild in guilds:
            guild_id = get_guild_id(guild)
            gateway = self.gateway_for(guild_id)
            
            try:
                guild_ids_for_gateway = by_gateway[gateway]
            except KeyError:
                guild_ids_for_gateway = []
                by_gateway[gateway] = guild_ids_for_gateway
            
            guild_ids_for_gateway.append(guild_id)
        
        
        for gateway, guild_ids_for_gateway in by_gateway.items():
            await gateway.send_as_json({
                'op': GATEWAY_OPERATION_CLIENT_REQUEST_SOUNDBOARD_SOUNDS,
                'd': {
                    'guild_ids': guild_ids_for_gateway,
                }
            })
        
        return await self.events.soundboard_sounds.wait_for_events_in_guilds(guild_ids)
    
    
    async def request_all_users_of(self, guild):
        """
        Requests all users of the given guild.
        
        > This method uses the client's gateway to request the users. If any of the parameters do not match their
        > expected value or if timeout occurs, returns instead of raising.
        
        If the ``Client`` is created with `should_request_users` parameter given as `False`, this parameter can be used
        to request users of specific guilds.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, what's members will be requested.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` or `int`.
        """
        guild_id = get_guild_id(guild)
        await self._request_users(guild_id)
    
    
    async def _request_users(self, guild_id):
        """
        Requests the members of the given guild. Called when the client joins a guild and user caching is enabled
        (so by default).
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild_id : ``int``
            The guild, what's members will be requested.
        """
        event_handler = self.events.guild_user_chunk
        
        self._user_chunker_nonce = nonce = self._user_chunker_nonce + 1
        nonce = format(nonce, '0>16x')
        
        event_handler.waiters[nonce] = waiter = MassUserChunker()
        
        data = {
            'op': GATEWAY_OPERATION_CLIENT_REQUEST_GUILD_USERS,
            'd': {
                'guild_id': guild_id,
                'query': '',
                'limit': 0,
                'presences': CACHE_PRESENCE,
                'nonce': nonce,
            },
        }
        
        gateway = self.gateway_for(guild_id)
        await gateway.send_as_json(data)
        
        if not (await waiter):
            try:
                del event_handler.waiters[nonce]
            except KeyError:
                pass
    
    
    async def request_users(self, guild, name, limit = 1):
        """
        Requests the members of the given guild by their name.
        
        > This method uses the client's gateway to request the users. If any of the parameters do not match their
        > expected value or if timeout occurs, returns an empty list instead of raising.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, what's members will be requested.
        name : `str`
            The received user's name or nick should start with this string.
        limit : `int` = `1`, Optional
            The amount of users to be received. Limited to `100`.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` or `int`.
        """
        guild_id = get_guild_id(guild)
        
        assert _assert__request_users__limit(limit)
        assert _assert__request_users__name(name)
        
        event_handler = self.events.guild_user_chunk
        
        self._user_chunker_nonce = nonce = self._user_chunker_nonce + 1
        nonce = format(nonce, '0>16x')
        
        event_handler.waiters[nonce] = waiter = SingleUserChunker()
        
        data = {
            'op': GATEWAY_OPERATION_CLIENT_REQUEST_GUILD_USERS,
            'd': {
                'guild_id': guild_id,
                'query': name,
                'limit': limit,
                'presences': CACHE_PRESENCE,
                'nonce': nonce,
            },
        }
        
        gateway = self.gateway_for(guild_id)
        await gateway.send_as_json(data)
        
        users = await waiter
        if users is None:
            try:
                del event_handler.waiters[nonce]
            except KeyError:
                pass
            
            users = []
        
        return users
    
    
    async def wait_for(self, event_name, check, timeout = None):
        """
        O(n) event waiter with massive overhead compared to other optimized event waiters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        event_name : `str`
            The respective event's name.
        check : `callable`
            Check, what tells that the waiting is over.
            
            If the `check` returns `True` the received `args` are passed to the waiter future and returned by the
            method. However if the check returns any non `bool` value, then that object is passed next to `args` and
            returned as well.
        
        timeout : `None`, `float` = `None`, Optional
            Timeout after `TimeoutError` is raised and the waiting is cancelled.
        
        Returns
        -------
        result : `object`
            Parameters passed to the `check` and the value returned by the `check` if it's type is not `bool`.
        
        Raised
        ------
        TimeoutError
            Timeout occurred.
        BaseException
            Any exception raised by `check`.
        """
        wait_for_handler = self.events.get_handler(event_name, WaitForHandler)
        if wait_for_handler is None:
            wait_for_handler = WaitForHandler()
            self.events(wait_for_handler, name = event_name)
        
        future = Future(KOKORO)
        wait_for_handler.waiters[future] = check
        
        if (timeout is not None):
            future.apply_timeout(timeout)
        
        try:
            return await future
        finally:
            waiters = wait_for_handler.waiters
            del waiters[future]
            
            if not waiters:
                self.events.remove(wait_for_handler, name = event_name)



__all__ = ()

from .fields import (
    put_cancelled, put_end, put_start, put_timestamp, validate_cancelled, validate_end, validate_start,
    validate_timestamp
)


SCHEDULED_EVENT_OCCASION_OVERWRITE_CREATE_FIELD_CONVERTERS = {
    'cancelled': (validate_cancelled, put_cancelled),
    'end': (validate_end, put_end),
    'start': (validate_start, put_start),
    'timestamp': (validate_timestamp, put_timestamp),
}

SCHEDULED_EVENT_OCCASION_OVERWRITE_EDIT_FIELD_CONVERTERS = SCHEDULED_EVENT_OCCASION_OVERWRITE_CREATE_FIELD_CONVERTERS



__all__ = ()

from datetime import datetime as DateTime, timedelta as TimeDelta, timezone as TimeZone

from scarletio import include_with_callback, set_docs

from .bases import maybe_snowflake
from .preconverters import preconvert_int_options, preconvert_preinstanced_type, preconvert_snowflake, preconvert_str
from .utils import is_url


ZERO_TIMEDELTA = TimeDelta()


def field_validator_factory(field_name):
    """
    Returns an any field validator. Allows base types.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(field_value):
        """
        Validates the given field.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : object
            The field value to validate.
        
        Returns
        -------
        field_value : `object`
        """
        nonlocal field_name
        
        if (field_value is not None):
            if isinstance(field_value, str):
                if not field_value:
                    field_value = None
            
            elif isinstance(field_value, (int, float)):
                pass
            
            else:
                raise TypeError(
                    f'`{field_name}` can be any basic type, like `None`, `str`, `int`, `float`, got '
                    f'{field_value.__class__.__name__}; {field_value!r}.'
                )
            
        return field_value
    
    return validator


def entity_id_validator_factory(field_name, entity_type = None, *, include = None):
    """
    Returns an entity id validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    entity_type : `None`, ``DiscordEntity`` = `None`, Optional
        The accepted entity's type.
    
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `FunctionType`
    """
    if (entity_type is None) and (include is None):
        def validator(entity_id):
            """
            Validates the given entity identifier.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_id : `int`
                The entity's identifier.
            
            Returns
            -------
            entity_id : `int`
                The entity's identifier.
            
            Raises
            ------
            TypeError
                - If `entity_id`'s type is incorrect.
            """
            return preconvert_snowflake(entity_id, field_name)
    
    else:
        def validator(entity_id):
            """
            Validates the given entity identifier.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_id : `None`, `entity_type`, `int`
                The entity or it's identifier.
            
            Returns
            -------
            entity_id : `int`
                The entity's identifier.
            
            Raises
            ------
            TypeError
                - If `entity_id`'s type is incorrect.
            """
            nonlocal field_name
            nonlocal entity_type
            
            if entity_id is None:
                processed_entity_id = 0
            
            elif isinstance(entity_id, entity_type):
                processed_entity_id = entity_id.id
            
            else:
                processed_entity_id = maybe_snowflake(entity_id)
                if processed_entity_id is None:
                    raise TypeError(
                        f'`{field_name}` can be `int`, `{entity_type.__name__}`, `int`, got '
                        f'{type(entity_id).__name__}; {entity_id!r}.'
                    )
            
            return processed_entity_id
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_entity_type(value):
            nonlocal entity_type
            entity_type = value
    
    
    return validator


def _entity_id_array_processor_factory(field_name, entity_type, include, keep_order):
    """
    Returns an entity id array validator. The validator does no `None` check, instead that should be done before.
    The validator always returns a set.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    entity_type : `type` with `{id}`
        The accepted entity's type.
    
    include : `None`, `str`
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    keep_order : `bool`
        Whether order should be kept. if given as `True` returns validator returning a `list` else a `set`.
    """
    if keep_order:
        if (entity_type is None):
            def validator(entity_id_array):
                nonlocal field_name
                
                if (getattr(entity_id_array, '__iter__', None) is None):
                    raise TypeError(
                        f'`{field_name}` can be `None | iterable<int>`, '
                        f'got {type(entity_id_array).__name__}; {entity_id_array!r}.'
                    )
                
                entity_id_set_processed = None
                
                for applied_tag_id in entity_id_array:
                    applied_tag_id_processed = maybe_snowflake(applied_tag_id)
                    if applied_tag_id_processed is None:
                        raise TypeError(
                            f'`{field_name}` can contain `int` elements, got {type(applied_tag_id).__name__}'
                            f'{applied_tag_id!r}; entity_id_array = {entity_id_array!r}.'
                        )
                    
                    if entity_id_set_processed is None:
                        entity_id_set_processed = []
                    
                    elif applied_tag_id_processed in entity_id_set_processed:
                        continue
                    
                    entity_id_set_processed.append(applied_tag_id_processed)
                
                return entity_id_set_processed
        else:
            def validator(entity_id_array):
                nonlocal field_name
                nonlocal entity_type
                
                if (getattr(entity_id_array, '__iter__', None) is None):
                    raise TypeError(
                        f'`{field_name}` can be `None`, `iterable` of (`int`, `{entity_type.__name__}`), '
                        f'got {type(entity_id_array).__name__}; {entity_id_array!r}.'
                    )
                
                entity_id_set_processed = None
                
                for applied_tag_id in entity_id_array:
                    if isinstance(applied_tag_id, entity_type):
                         applied_tag_id_processed = applied_tag_id.id
                    
                    else:
                        applied_tag_id_processed = maybe_snowflake(applied_tag_id)
                        if applied_tag_id_processed is None:
                            raise TypeError(
                                f'`{field_name}` can contain `int`, `{entity_type.__name__}` elements, got '
                                f'{type(applied_tag_id).__name__}; {applied_tag_id!r}; '
                                f'entity_id_array = {entity_id_array!r}.'
                            )
                    
                    if entity_id_set_processed is None:
                        entity_id_set_processed = []
                    
                    elif applied_tag_id_processed in entity_id_set_processed:
                        continue
                    
                    entity_id_set_processed.append(applied_tag_id_processed)
                
                return entity_id_set_processed
    
    else:
        if (entity_type is None):
            def validator(entity_id_array):
                nonlocal field_name
                
                if (getattr(entity_id_array, '__iter__', None) is None):
                    raise TypeError(
                        f'`{field_name}` can be `None | iterable<int>`, '
                        f'got {type(entity_id_array).__name__}; {entity_id_array!r}.'
                    )
                
                entity_id_set_processed = None
                
                for applied_tag_id in entity_id_array:
                    applied_tag_id_processed = maybe_snowflake(applied_tag_id)
                    if applied_tag_id_processed is None:
                        raise TypeError(
                            f'`{field_name}` can contain `int` elements, got {type(applied_tag_id).__name__}'
                            f'{applied_tag_id!r}; entity_id_array = {entity_id_array!r}.'
                        )
                    
                    if entity_id_set_processed is None:
                        entity_id_set_processed = set()
                    
                    entity_id_set_processed.add(applied_tag_id_processed)
                
                return entity_id_set_processed
        else:
            def validator(entity_id_array):
                nonlocal field_name
                nonlocal entity_type
                
                if (getattr(entity_id_array, '__iter__', None) is None):
                    raise TypeError(
                        f'`{field_name}` can be `None`, `iterable` of (`int`, `{entity_type.__name__}`), '
                        f'got {type(entity_id_array).__name__}; {entity_id_array!r}.'
                    )
                
                entity_id_set_processed = None
                
                for applied_tag_id in entity_id_array:
                    if isinstance(applied_tag_id, entity_type):
                         applied_tag_id_processed = applied_tag_id.id
                    
                    else:
                        applied_tag_id_processed = maybe_snowflake(applied_tag_id)
                        if applied_tag_id_processed is None:
                            raise TypeError(
                                f'`{field_name}` can contain `int`, `{entity_type.__name__}` elements, got '
                                f'{type(applied_tag_id).__name__}; {applied_tag_id!r}; '
                                f'entity_id_array = {entity_id_array!r}.'
                            )
                    
                    if entity_id_set_processed is None:
                        entity_id_set_processed = set()
                    
                    entity_id_set_processed.add(applied_tag_id_processed)
                
                return entity_id_set_processed
    
    if (include is not None):
        @include_with_callback(include)
        def include_entity_type(value):
            nonlocal entity_type
            entity_type = value
    
    
    return validator


def entity_id_array_validator_factory(field_name, entity_type = None, *, include = None, ordered = True):
    """
    Returns an entity id array validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    entity_type : `type` with `{id}` = `None`, Optional
        The accepted entity's type.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    ordered : `bool` = `True`, Optional (Keyword only)
        Whether the output should be ordered.
    
    Returns
    -------
    validator : `FunctionType`
    """
    base_validator = _entity_id_array_processor_factory(field_name, entity_type, include, not ordered)
    if ordered:
        def validator(entity_id_array):
            nonlocal base_validator
            
            if entity_id_array is None:
                return None
            
            entity_ids_processed = base_validator(entity_id_array)
            if (entity_ids_processed is not None):
                return tuple(sorted(entity_ids_processed))
    
    else:
        def validator(entity_id_array):
            nonlocal base_validator
            
            if entity_id_array is None:
                return None
            
            entity_ids_processed = base_validator(entity_id_array)
            if (entity_ids_processed is not None):
                return tuple(entity_ids_processed)
    
    set_docs(
        validator,
        """
        Validates the given entity identifier array.
        
        > This function is generated.
        
        Parameters
        ----------
        entity_id_array : `None`, `iterable` of (`int`, `entity_type`) items
            The entities or their identifiers.
        
        Returns
        -------
        entity_id_array : `None | tuple<int>`
            The entities identifiers.
        
        Raises
        ------
        TypeError
            - If `entity_id_array`'s type is incorrect.
            - If an element of `entity_id_array` has incorrect type.
        """
    )
    
    return validator


def entity_id_set_validator_factory(field_name, entity_type, *, include = None):
    """
    Returns an entity id set validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    entity_type : `type` with `{id}`
        The accepted entity's type.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `FunctionType`
    """
    base_validator = _entity_id_array_processor_factory(field_name, entity_type, include, False)
    
    def validator(entity_id_array):
        """
        Validates the given entity identifier array.
        
        > This function is generated.
        
        Parameters
        ----------
        entity_id_array : `None`, `iterable` of (`int`, `entity_type`) items
            The entities or their identifiers.
        
        Returns
        -------
        entity_id_set : `tuple` of `int`
            The entities identifiers.
        
        Raises
        ------
        TypeError
            - If `entity_id_array`'s type is incorrect.
            - If an element of `entity_id_array` has incorrect type.
        """
        nonlocal base_validator
        
        if entity_id_array is None:
            return set()
        
        entity_ids_processed = base_validator(entity_id_array)
        if (entity_ids_processed is None):
            return set()
        
        return entity_ids_processed
    
    return validator


def preinstanced_validator_factory(field_name, preinstanced_type, *, include = None):
    """
    Returns a preinstanced validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    preinstanced_type : ``PreinstancedBase``
        The accepted preinstanced type.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The type's name to include `preinstanced_type` with.
        Should be used when `preinstanced_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(preinstanced):
        """
        Validates the given preinstanced object.
        
        > This function is generated.
        
        Parameters
        ----------
        preinstanced : ``PreinstancedBase``, `instance<PreinstancedBase.VALUE_TYPE>`
            The preinstanced object or it's value.
        
        Returns
        -------
        preinstanced : ``PreinstancedBase``
        
        Raises
        ------
        TypeError
            - If `preinstanced`'s type is incorrect.
        """
        nonlocal field_name
        nonlocal preinstanced_type
        
        return preconvert_preinstanced_type(preinstanced, field_name, preinstanced_type)
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal preinstanced_type
            preinstanced_type = value
    
    return validator


def nullable_preinstanced_validator_factory(field_name, preinstanced_type, default):
    """
    Returns a nullable preinstanced validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    preinstanced_type : ``PreinstancedBase``
        The accepted preinstanced type.
    default : `instance<preinstanced_type>`
        The default value to use if `None` is given.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(preinstanced):
        """
        Validates the given preinstanced object.
        
        > This function is generated.
        
        Parameters
        ----------
        preinstanced : `None`, ``PreinstancedBase``, `instance<PreinstancedBase.VALUE_TYPE>`
            The preinstanced object or it's value.
        
        Returns
        -------
        preinstanced : ``PreinstancedBase``
        
        Raises
        ------
        TypeError
            - If `preinstanced`'s type is incorrect.
        """
        nonlocal default
        nonlocal field_name
        nonlocal preinstanced_type
        
        if preinstanced is None:
            return default
        
        return preconvert_preinstanced_type(preinstanced, field_name, preinstanced_type)
    
    return validator


def preinstanced_array_validator_factory(field_name, preinstanced_type, *, include = None):
    """
    Returns a preinstanced array validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    preinstanced_type : ``PreinstancedBase``
        The accepted preinstanced type.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The type's name to include `preinstanced_type` with.
        Should be used when `preinstanced_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(preinstanced_array):
        """
        Validates the given preinstanced object.
        
        > This function is generated.
        
        Parameters
        ----------
        preinstanced_array : `None`, `iterable` of (``PreinstancedBase``, `instance<PreinstancedBase.VALUE_TYPE>`)
            The preinstanced objects or their value.
        
        Returns
        -------
        preinstanced_array : `None`, `tuple` of ``PreinstancedBase``
        
        Raises
        ------
        TypeError
            - If `preinstanced`'s type is incorrect.
        """
        nonlocal field_name
        nonlocal preinstanced_type
        
        if preinstanced_array is None:
            return None
        
        if isinstance(preinstanced_array, preinstanced_type):
            return (preinstanced_array, )
        
        if isinstance(preinstanced_array, preinstanced_type.VALUE_TYPE):
            return (preinstanced_type(preinstanced_array), )
        
        if getattr(preinstanced_array, '__iter__', None) is None:
            raise TypeError(
                f'`{field_name}` can be `None`, `{preinstanced_type.__name__}`, '
                f'`{preinstanced_type.VALUE_TYPE.__name__}`, or `iterable`, got '
                f'{type(preinstanced_array).__name__}; {preinstanced_array!r}.'
            )
        
        unique_elements = None
        
        for preinstanced in preinstanced_array:
            if isinstance(preinstanced, preinstanced_type):
                pass
            
            elif isinstance(preinstanced, preinstanced_type.VALUE_TYPE):
                preinstanced = preinstanced_type(preinstanced)
            
            else:
                raise TypeError(
                    f'`{field_name}` elements can be `{preinstanced_type.__name__}`, '
                    f'`{preinstanced_type.VALUE_TYPE.__name__}`, '
                    f'got {type(preinstanced).__name__}; {preinstanced!r}.'
                )
            
            if unique_elements is None:
                unique_elements = set()
            
            unique_elements.add(preinstanced)
        
        if unique_elements is None:
            return None
        
        return tuple(sorted(unique_elements))
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal preinstanced_type
            preinstanced_type = value
    
    return validator


def _field_condition_validator_factory(field_name, field_type, default_value, condition_check, condition_message):
    """
    Returns a new field validator with an added condition validation.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    field_type : `type`
        The field's type to only accept.
    default_value : `object`
        The default to return.
    condition_check : `callable`
        The condition which needs to pass.
    condition_message : `str`
        Condition message to use when building the error message.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(value):
        """
        Validates the given field value.
        
        > This function is generated.
        
        Parameters
        ----------
        value : `object`
            The field to validate.
        
        Returns
        -------
        value : `object`
        
        Raises
        ------
        TypeError
            - If `field` is not `field_type`.
        ValueError
            - If `condition_check` fails for the given value.
        """
        nonlocal condition_check
        nonlocal default_value
        nonlocal condition_message
        nonlocal field_name
        nonlocal field_type
        
        if value is None:
            value = default_value
        
        else:
            if not isinstance(value, field_type):
                raise TypeError(
                    f'`{field_name}` can be `{field_type.__name__}`, got {value.__class__.__name__}; {value!r}.'
                )
            
            if not condition_check(value):
                raise ValueError(
                    f'`{field_name}` must be {condition_message}, got {value!r}.'
                )
        
        return value
    
    return validator


def int_conditional_validator_factory(field_name, default_value, condition_check, condition_message):
    """
    Returns a new `int` with condition validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    default_value : `int`
        The default to return.
    condition_check : `callable`
        The condition which needs to pass.
    condition_message : `str`
        Condition message to use when building the error message.
    
    Returns
    -------
    validator : `FunctionType`
    """
    return _field_condition_validator_factory(field_name, int, default_value, condition_check, condition_message)


def float_conditional_validator_factory(field_name, default_value, condition_check, condition_message):
    """
    Returns a new `float` with condition validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    default_value : `float`
        The default to return.
    condition_check : `callable`
        The condition which needs to pass.
    condition_message : `str`
        Condition message to use when building the error message.
    
    Returns
    -------
    validator : `FunctionType`
    """
    return _field_condition_validator_factory(field_name, float, default_value, condition_check, condition_message)


def flag_validator_factory(field_name, flag_type, *, default_value = ...):
    """
    Returns a new flag validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    flag_type : `type`
        The flag type to use.
    default_value : `instance<flag_type>`, Optional (Keyword only)
        Default value to return if value is given as `None`.
    
    Returns
    -------
    validator : `FunctionType`
    """
    if default_value is ...:
        default_value = flag_type()
    
    def validator(flag):
        """
        Validates the given flag.
        
        > This function is generated.
        
        Parameters
        ----------
        flag : `None | int | instance<flag_type>`
            The flag to validate.
        
        Returns
        -------
        integer : `instance<flag_type>`
        
        Raises
        ------
        TypeError
            - If `flag` is not `int`.
        ValueError
            - If `flag` is not any of the expected options.
        """
        nonlocal default_value
        nonlocal field_name
        nonlocal flag_type
        
        if flag is None:
            flag = default_value
        
        elif isinstance(flag, flag_type):
            pass
        
        elif isinstance(flag, int):
            flag = flag_type(flag)
        
        else:
            raise TypeError(
                f'`{field_name}` can be `None | int | {flag_type.__name__}`, '
                f'got {type(flag).__name__}; {flag!r}.'
            )
        
        return flag
    
    return validator


def nullable_flag_validator_factory(field_name, flag_type):
    """
    Returns a new nullable flag validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    flag_type : `type`
        The flag type to use.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(flag):
        """
        Validates the given flag.
        
        > This function is generated.
        
        Parameters
        ----------
        flag : `None | int | instance<flag_type>`
            The flag to validate.
        
        Returns
        -------
        integer : `instance<flag_type>`
        
        Raises
        ------
        TypeError
            - If `flag` is not `int`.
        ValueError
            - If `flag` is not any of the expected options.
        """
        nonlocal field_name
        nonlocal flag_type
        
        if flag is None:
            pass
        
        elif isinstance(flag, flag_type):
            pass
        
        elif isinstance(flag, int):
            flag = flag_type(flag)
        
        else:
            raise TypeError(
                f'`{field_name}` can be `None | int | {flag_type.__name__}`, '
                f'got {flag.__class__.__name__}; {flag!r}.'
            )
    
        return flag
    
    return validator


def int_options_validator_factory(field_name, field_options, default_value):
    """
    Returns a `int` with options validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    field_options : `frozenset<int>`
        The allowed values.
    
    default_value : `int`
        The default value to return if `None` is received.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(integer):
        """
        Validates the given integer value.
        
        > This function is generated.
        
        Parameters
        ----------
        integer : `int`
            The integer to validate.
        
        Returns
        -------
        integer : `int`
        
        Raises
        ------
        TypeError
            - If `integer` is not `int`.
        ValueError
            - If `integer` is not any of the expected options.
        """
        nonlocal default_value
        nonlocal field_name
        nonlocal field_options
        
        if integer is None:
            return default_value
        
        return preconvert_int_options(integer, field_name, field_options)
    
    return validator


def bool_validator_factory(field_name, default_value):
    """
    Returns a new `bool` validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    default_value : `bool`
        The default value to return if `None` is received.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(boolean):
        """
        Validates the given boolean value.
        
        > This function is generated.
        
        Parameters
        ----------
        boolean : `None`, `bool`, `int`
            The boolean to validate.
        
        Returns
        -------
        boolean : `int`
        
        Raises
        ------
        TypeError
            - If `boolean`'s type is incorrect.
        ValueError
            - If `boolean`'s value is incorrect.
        """
        nonlocal field_name
        nonlocal default_value
        
        if type(boolean) is bool:
            return boolean
        
        if boolean is None:
            return default_value
        
        if isinstance(boolean, int):
            if (boolean not in (0, 1)):
                raise ValueError(
                    f'`{field_name}` was given as `int`, but neither as `0`, `1`, got {boolean!r}.'
                )
            
            return bool(boolean)
        
        raise TypeError(
            f'`{field_name}` can be `None`, `bool`, `int` as `0`, `1`, got {boolean.__class__.__name__}; {boolean!r}.'
        )
    
    return validator


def nullable_date_time_validator_factory(field_name):
    """
    Returns a new nullable `DateTime` validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(date_time):
        """
        Validates the given date time value.
        
        > This function is generated.
        
        Parameters
        ----------
        date_time : `None | DateTime`
            The date time to validate.
        
        Returns
        -------
        date_time : `None | DateTime`
        
        Raises
        ------
        TypeError
            - If `date_time` is not `None | DateTime`.
        """
        nonlocal field_name
        
        if (date_time is not None) and (not isinstance(date_time, DateTime)):
            raise TypeError(
                f'`{field_name} can be `None`, `{DateTime.__name__}`, '
                f'got {date_time.__class__.__name__}; {date_time!r}.'
            )
        
        return date_time
    
    return validator


def force_date_time_validator_factory(field_name):
    """
    Returns a new non-nullable `DateTime` validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(date_time):
        """
        Validates the given date time value.
        
        > This function is generated.
        
        Parameters
        ----------
        date_time : `DateTime`
            The date time to validate.
        
        Returns
        -------
        date_time : `DateTime`
        
        Raises
        ------
        TypeError
            - If `date_time` is not `DateTime`.
        """
        nonlocal field_name
        
        if (not isinstance(date_time, DateTime)):
            raise TypeError(
                f'`{field_name} can be `{DateTime.__name__}`, got {date_time.__class__.__name__}; {date_time!r}.'
            )
        
        return date_time
    
    return validator


def force_string_validator_factory(field_name, length_min, length_max):
    """
    Returns an enforced string validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    length_min : `int`
        The minimal allowed string length.
    length_max : `int`
        The maximal allowed string length.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(string):
        """
        Validates the given string.
        
        > This function is generated.
        
        Parameters
        ----------
        string : `None`, `str`
            The string to validate.
        
        Returns
        -------
        string : `str`
                
        Raises
        ------
        TypeError
            - If `string` is not `None`, `str`.
        ValueError
            - If `string`'s length is out of the expected range.
        """
        nonlocal field_name
        nonlocal length_min
        nonlocal length_max
        
        if (string is None):
            string = ''
        else:
            string = preconvert_str(string, field_name, length_min, length_max)
        
        return string
    
    return validator


def nullable_string_validator_factory(field_name, length_min, length_max):
    """
    Returns a nullable string validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    length_min : `int`
        The minimal allowed string length.
    length_max : `int`
        The maximal allowed string length.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(string):
        """
        Validates the given string.
        
        > This function is generated.
        
        Parameters
        ----------
        string : `None`, `str`
            The string to validate.
        
        Returns
        -------
        string : `None`, `str`
                
        Raises
        ------
        TypeError
            - If `string` is not `None`, `str`.
        ValueError
            - If `string`'s length is out of the expected range.
        """
        nonlocal field_name
        nonlocal length_min
        nonlocal length_max
        
        if (string is not None):
            string = preconvert_str(string, field_name, length_min, length_max)
            if (not string):
                string = None
        
        return string
    
    return validator


def nullable_string_array_validator_factory(field_name, *, ordered = True):
    """
    Returns a nullable string array validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    ordered : `bool` = `True`, Optional (Keyword only)
        Whether the output should be ordered.
    
    Returns
    -------
    validator : `FunctionType`
    """
    if ordered:
        def validator(string_array):
            """
            Validates the given string array. Returns the elements in order.
            
            > This function is generated.
            
            Parameters
            ----------
            string_array : `None | str | iterable<str>`
                The string to validate.
            
            Returns
            -------
            string_array : `None | tuple<str>
                    
            Raises
            ------
            TypeError
                - If `string_array`'s type is incorrect.
            """
            nonlocal field_name
            
            if (string_array is None):
                return None
            
            if isinstance(string_array, str):
                return (string_array, )
            
            if getattr(string_array, '__iter__', None) is None:
                raise TypeError(
                    f'`{field_name}` can be `None`, `iterable` of `str`, got '
                    f'{string_array.__class__.__name__}; {string_array!r}.'
                )
            
            processed_values = None
            
            for string in string_array:
                if not isinstance(string, str):
                    raise TypeError(
                        f'`{field_name}` elements can be `str`, got '
                        f'{type(string).__name__}; {string!r}; {field_name} = {string_array!r}.'
                    )
                
                if (processed_values is None):
                    processed_values = set()
                
                processed_values.add(string)
            
            if processed_values is not None:
                return tuple(sorted(processed_values))
    
    else:
        def validator(string_array):
            """
            Validates the given string array. Returns the elements in order as they are present in the input.
            
            > This function is generated.
            
            Parameters
            ----------
            string_array : `None | str, iterable<str>`
                The string to validate.
            
            Returns
            -------
            string_array : `None | tuple<str>`
                    
            Raises
            ------
            TypeError
                - If `string_array`'s type is incorrect.
            """
            nonlocal field_name
            
            if (string_array is None):
                return None
            
            if isinstance(string_array, str):
                return (string_array, )
            
            if getattr(string_array, '__iter__', None) is None:
                raise TypeError(
                    f'`{field_name}` can be `None`, `iterable` of `str`, got '
                    f'{type(string_array).__name__}; {string_array!r}.'
                )
            
            processed_values = None
            
            for string in string_array:
                if not isinstance(string, str):
                    raise TypeError(
                        f'`{field_name}` elements can be `str`, got '
                        f'{type(string).__name__}; {string!r}; {field_name} = {string_array!r}.'
                    )
                
                if (processed_values is None):
                    processed_values = []
                
                processed_values.append(string)
            
            if processed_values is not None:
                return tuple(processed_values)
    
    
    return validator



def nullable_sorted_int_array_conditional_validator_factory(field_name, condition_check, condition_message):
    """
    Returns a nullable sorted int array validator with
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    condition_check : `callable`
        The condition which needs to pass.
    condition_message : `str`
        Condition message to use when building the error message.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(int_array):
        """
        Validates the given int array. Returns the elements in order.
        
        > This function is generated.
        
        Parameters
        ----------
        int_array : `None | int | iterable<int>`
            The string to validate.
        
        Returns
        -------
        int_array : `None | tuple<int>`
                
        Raises
        ------
        TypeError
            - If `int_array`'s type is incorrect.
            - If a value in `int_array` is incorrect.
        """
        nonlocal field_name
        nonlocal condition_check
        nonlocal condition_message
        
        if (int_array is None):
            return None
        
        if isinstance(int_array, int):
            if not condition_check(int_array):
                raise ValueError(
                    f'`{field_name}` must be {condition_message}, got {int_array!r}.'
                )
            
            return (int_array, )
        
        if getattr(int_array, '__iter__', None) is None:
            raise TypeError(
                f'`{field_name}` can be `None | iterable<int>`, got '
                f'{type(int_array).__name__}; {int_array!r}.'
            )
        
        processed_values = None
        
        for int_value in int_array:
            if not isinstance(int_value, int):
                raise TypeError(
                    f'`{field_name}` elements can be `int`, got '
                    f'{type(int_array).__name__}; {int_value!r}; {field_name} = {int_value!r}.'
                )
            
            if not condition_check(int_value):
                raise ValueError(
                    f'`{field_name}` must be {condition_message}, got {int_value!r}.'
                )
            
            if (processed_values is None):
                processed_values = set()
            
            processed_values.add(int_value)
        
        if processed_values is not None:
            return tuple(sorted(processed_values))
    
    return validator


def url_optional_validator_factory(field_name, *, length_max = ...):
    """
    Returns an optional url validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    length_max : `int`, Optional (Keyword only)
        The maximal allowed length of the url.
    
    Returns
    -------
    validator : `FunctionType`
    """
    if length_max is ...:
        length_max = 2048
    
    def validator(url):
        """
        Validates the given string.
        
        > This function is generated.
        
        Parameters
        ----------
        url : `None`, `str`
            The url to validate.
        
        Returns
        -------
        string : `None`, `str`
                
        Raises
        ------
        TypeError
            - If `url` is not `None`, `str`.
        ValueError
            - If `url` is not an url.
        """
        nonlocal field_name
        nonlocal length_max
        
        if (url is not None):
            if not isinstance(url, str):
                raise TypeError(
                    f'`{field_name!s}` can be `None`, `str`, got {url.__class__.__name__!s}; {url!r}.'
                )
            
            url_length = len(url)
            if url_length:
                if url_length > length_max:
                    raise ValueError(
                        f'`{field_name!s}` can be up to {length_max!s} character long, got {url_length!s}; {url!r}.'
                    )
                    
                if not is_url(url):
                    raise ValueError(
                        f'`{field_name!s}` is not a valid url, got {url!r}.'
                    )
                
            else:
                url = None
        
        return url
    
    return validator


def url_required_validator_factory(field_name):
    """
    Returns a required url validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(url):
        """
        Validates the given string.
        
        > This function is generated.
        
        Parameters
        ----------
        url : `None`, `str`
            The url to validate.
        
        Returns
        -------
        string : `None`, `str`
                
        Raises
        ------
        TypeError
            - If `url` is not `None`, `str`.
        ValueError
            - If `url` is not an url.
        """
        nonlocal field_name
        
        if (url is not None) and not isinstance(url, str):
            if not isinstance(url, str):
                raise TypeError(
                    f'`{field_name}` can be `None`, `str`, got {url.__class__.__name__}; {url!r}.'
                )
        
        if (url is None) or (not url):
            if not isinstance(url, str):
                raise ValueError(
                    f'`{field_name}` cannot be empty, got {type(url).__name__}; {url!r}.'
                )
        
        if not is_url(url):
            raise ValueError(
                f'`{field_name}` is not a valid url, got {url!r}.'
            )
        
        return url
    
    return validator


def url_array_optional_validator_factory(field_name):
    """
    Returns an optional url array validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(url_array):
        """
        Validates the given string.
        
        > This function is generated.
        
        Parameters
        ----------
        url_array : `None | str | iterable<str>`
            The url array to validate.
        
        Returns
        -------
        url_array : `None`, `str`
                
        Raises
        ------
        TypeError
            - If `url` is not `None | str | iterable<str>`
        ValueError
            - If an `url` is not an url.
        """
        nonlocal field_name
        
        if (url_array is None):
            return None
        
        if isinstance(url_array, str):
            if not is_url(url_array):
                raise ValueError(
                    f'`{field_name}` is not a valid url, got {url_array!r}.'
                )
            
            return (url_array, )
        
        if getattr(url_array, '__iter__', None) is None:
            raise TypeError(
                f'`{field_name}` can be `None`, `iterable` of `str`, got '
                f'{type(url_array).__name__}; {url_array!r}.'
            )
        
        processed_values = None
        
        for url in url_array:
            if not isinstance(url, str):
                raise TypeError(
                    f'`{field_name}` elements can be `str`, got '
                    f'{type(url).__name__}; {url!r}; {field_name} = {url_array!r}.'
                )
            
            if not is_url(url):
                raise ValueError(
                    f'`{field_name}` element is not a valid url, got {url!r}; {field_name} = {url_array!r}.'
                )
            
            if (processed_values is None):
                processed_values = set()
            
            processed_values.add(url)
        
        if processed_values is not None:
            return tuple(sorted(processed_values))
    
    return validator


def nullable_entity_set_validator_factory(field_name, entity_type, *, include = None):
    """
    Returns a nullable entity array validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    entity_type : `type`
        The allowed entity type.
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(entity_array):
        """
        Validates the given nullable entity array field.
        
        Parameters
        ----------
        entity_array : `None`, `iterable` of `instance<entity_type>`
            The entity array to validate.
        
        Returns
        -------
        entity_array : `None`, `tuple` of `instance<entity_type>`
        
        Raises
        ------
        TypeError
            - If `entity_array` is not `None`, `iterable` of `entity_type`.
        """
        nonlocal field_name
        nonlocal entity_type
        
        if entity_array is None:
            return None
        
        if (getattr(entity_array, '__iter__', None) is None):
            raise TypeError(
                f'`{field_name}` can be `None`, `iterable` of `{entity_type.__name__}`, got '
                f'{entity_array.__class__.__name__}; {entity_array!r}.'
            )
            
        entity_set_processed = None
        
        for entity in entity_array:
            if not isinstance(entity, entity_type):
                raise TypeError(
                    f'`{field_name}` can contain `{entity_type.__name__}` elements, got '
                    f'{entity.__class__.__name__}; {entity!r}; {field_name!s} = {entity_array!r}.'
                )
            
            if (entity_set_processed is None):
                entity_set_processed = set()
            
            entity_set_processed.add(entity)
        
        return entity_set_processed
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal entity_type
            entity_type = value
    
    return validator


def nullable_entity_array_validator_factory(field_name, entity_type, *, include = None, sort_key = None):
    """
    Returns a nullable entity array validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    entity_type : `type`
        The allowed entity type.
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    sort_key : `None`, `FunctionType` = `None`, Optional (Keyword only)
        Sort key used to sort the entities by.
    
    Returns
    -------
    validator : `FunctionType`
    """
    base_validator = nullable_entity_set_validator_factory(field_name, entity_type, include = include)
    
    def validator(entity_array):
        """
        Validates the given nullable entity array field.
        
        Parameters
        ----------
        entity_array : `None`, `iterable` of `instance<entity_type>`
            The entity array to validate.
        
        Returns
        -------
        entity_array : `None`, `tuple` of `instance<entity_type>`
        
        Raises
        ------
        TypeError
            - If `entity_array` is not `None`, `iterable` of `entity_type`.
        """
        nonlocal base_validator
        nonlocal sort_key
        
        entity_array_processed = base_validator(entity_array)
        if (entity_array_processed is not None):
            entity_array_processed = tuple(sorted(entity_array_processed, key = sort_key))
        
        return entity_array_processed
    
    return validator


def nullable_entity_validator_factory(field_name, entity_type, *, include = None):
    """
    Returns a nullable entity validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    entity_type : `type`
        The allowed entity type.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The entity type's name to include `entity_type` with.
        Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `FunctionType`
    """
    return default_entity_validator_factory(field_name, entity_type, default = None, include = include)


def nullable_entity_conditional_validator_factory(
    field_name, entity_type, condition_check, condition_message, *, include = None
):
    """
    Returns a nullable entity validator with a condition validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    entity_type : `type`
        The allowed entity type.
    
    condition_check : `callable`
        The condition which needs to pass.
    
    condition_message : `str`
        Condition message to use when building the error message.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(entity):
        """
        Validates the given nullable entity field.
        
        Parameters
        ----------
        entity : `None`, `instance<entity_type>`
            The entity to validate.
        
        Returns
        -------
        entity : `None`,  `instance<entity_type>`
        
        Raises
        ------
        TypeError
            - If `entity` is not `None`, ``entity_type``.
        """
        nonlocal field_name
        nonlocal entity_type
        nonlocal condition_check
        nonlocal condition_message
        
        if (entity is not None):
            if (not isinstance(entity, entity_type)):
                raise TypeError(
                    f'`{field_name}` can be `None`, `{entity_type.__name__}`, got '
                    f'{entity.__class__.__name__}; {entity!r}.'
                )
            
            if not condition_check(entity):
                raise ValueError(
                    f'`{field_name}` must be {condition_message}, got {entity!r}.'
                )
            
        return entity
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal entity_type
            entity_type = value
    
    
    return validator


def default_entity_validator_factory(field_name, entity_type, *, default = ..., default_factory = ..., include = None):
    """
    Returns a defaulted entity validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    entity_type : `type`
        The allowed entity type.
    
    default : `object`, Optional (Keyword only)
        The default value to return if `None` is received.
    
        Mutually exclusive with `default_factory` and at least one of them is required.
    
    default_factory : `() -> object`, Optional (Keyword only)
        Default value factory to call if `None` is received.
        
        Mutually exclusive with `default` and at least one of them is required.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The entity type's name to include `entity_type` with.
        Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `(object) -> instance<entity_type> | default | return<default_factory>`
    """
    if default is ... and default_factory is ...:
        raise TypeError(
            f'Either `default` or `default_factory` are required.'
        )
    
    if default is not ... and default_factory is not ...:
        raise TypeError(
            f'`default` and `default_factory` are mutually exclusive. '
            f'Got default = {default!r}, default_factory = {default_factory!r}.'
        )
    
    
    if default is not ...:
        def validator(entity):
            nonlocal default
            nonlocal field_name
            nonlocal entity_type
            
            if (entity is None):
                entity = default
            
            elif isinstance(entity, entity_type):
                pass
            
            else:
                raise TypeError(
                    f'`{field_name}` can be `None`, `{entity_type.__name__}`, got {entity.__class__.__name__}; {entity!r}.'
                )
            
            return entity
    
    else:
        def validator(entity):
            nonlocal default_factory
            nonlocal field_name
            nonlocal entity_type
            
            if (entity is None):
                entity = default_factory()
            
            elif isinstance(entity, entity_type):
                pass
            
            else:
                raise TypeError(
                    f'`{field_name}` can be `None`, `{entity_type.__name__}`, got {entity.__class__.__name__}; {entity!r}.'
                )
            
            return entity
    
    set_docs(
        validator,
        """
        Validates the given entity field value.
        
        Parameters
        ----------
        entity : `None`, `instance<entity_type>`
            The entity to validate.
        
        Returns
        -------
        entity : `None`,  `instance<entity_type>`
        
        Raises
        ------
        TypeError
            - If `entity` is not `None`, `entity_type`.
        """
    )
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal entity_type
            entity_type = value
    
    
    return validator


def entity_validator_factory(field_name, entity_type, *, include = None):
    """
    Returns an entity validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    entity_type : `type`
        The allowed entity type.
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(entity):
        """
        Validates the given nullable entity field.
        
        Parameters
        ----------
        entity : `instance<entity_type>`
            The entity to validate.
        
        Returns
        -------
        entity : `instance<entity_type>`
        
        Raises
        ------
        TypeError
            - If `entity` is not `None`, ``entity_type``.
        """
        nonlocal field_name
        nonlocal entity_type
        
        if not isinstance(entity, entity_type):
            raise TypeError(
                f'`{field_name}` can be `{entity_type.__name__}`, got {entity.__class__.__name__}; {entity!r}.'
            )
        
        return entity
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal entity_type
            entity_type = value
    
    return validator


def nullable_object_array_validator_factory(field_name, object_type, *, include = None, ordered = False):
    """
    Returns a nullable object array validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    object_type : `type`
        The allowed object type.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `object_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    validator : `FunctionType`
    """
    if ordered:
        def validator(object_array):
            nonlocal field_name
            nonlocal object_type
            
            if object_array is None:
                return None
            
            if (getattr(object_array, '__iter__', None) is None):
                raise TypeError(
                    f'`{field_name}` can be `None`, `iterable` of `{object_type.__name__}`, got '
                    f'{type(object_array).__name__}; {object_array!r}.'
                )
                
            object_array_processed = None
            
            for object in object_array:
                if not isinstance(object, object_type):
                    raise TypeError(
                        f'`{field_name}` can contain `{object_type.__name__}` elements, got '
                        f'{type(object).__name__}; {object!r}; object_array = {object_array!r}.'
                    )
                
                if (object_array_processed is None):
                    object_array_processed = []
                
                object_array_processed.append(object)
            
            if (object_array_processed is not None):
                object_array_processed.sort()
                object_array_processed = tuple(object_array_processed)
            
            return object_array_processed
    
    else:
        def validator(object_array):
            nonlocal field_name
            nonlocal object_type
            
            if object_array is None:
                return None
            
            if (getattr(object_array, '__iter__', None) is None):
                raise TypeError(
                    f'`{field_name}` can be `None`, `iterable` of `{object_type.__name__}`, got '
                    f'{type(object_array).__name__}; {object_array!r}.'
                )
                
            object_array_processed = None
            
            for object in object_array:
                if not isinstance(object, object_type):
                    raise TypeError(
                        f'`{field_name}` can contain `{object_type.__name__}` elements, got '
                        f'{type(object).__name__}; {object!r}; object_array = {object_array!r}.'
                    )
                
                if (object_array_processed is None):
                    object_array_processed = []
                
                object_array_processed.append(object)
            
            if (object_array_processed is not None):
                object_array_processed = tuple(object_array_processed)
            
            return object_array_processed
    
    set_docs(
        validator,
        """
        Validates the given nullable object array field.
        
        Parameters
        ----------
        object_array : `None`, `iterable` of `instance<object_type>`
            The object array to validate.
        
        Returns
        -------
        object_array : `None`, `tuple` of `instance<object_type>`
        
        Raises
        ------
        TypeError
            - If `object_array` is not `None`, `iterable` of `object_type`.
        """
    )
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal object_type
            object_type = value
    
    
    return validator


def entity_dictionary_validator_factory(field_name, entity_type):
    """
    Returns an entity dictionary validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    entity_type : `type<DiscordEntity>`
        The allowed entity type.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(field_value):
        """
        Validates the given entity dictionary field.
        
        Parameters
        ----------
        field_value : `None`, `iterable` of `instance<entity_type>`, `dict` of (`int`, `entity_type`) items
            The field value to validate.
        
        Returns
        -------
        entity_dictionary : `dict` of (`int`, `instance<entity_type>`) items
        
        Raises
        ------
        TypeError
            - If `field_value`'s or it's elements type is incorrect.
        """
        nonlocal field_name
        nonlocal entity_type
        
        validated = {}
        
        if field_value is not None:
            if isinstance(field_value, dict):
                iterator = iter(field_value.values())
            
            elif (getattr(field_value, '__iter__', None) is not None):
                iterator = iter(field_value)
            
            else:
                raise TypeError(
                    f'`{field_name}` can be `None`, `dict` of (`int`, `{entity_type.__name__}`) items, `iterable` of '
                    f'`{entity_type.__name__}` items, got {field_value.__class__.__name__}; {field_value!r}.'
                )
            
            for element in iterator:
                if not isinstance(element, entity_type):
                    raise TypeError(
                        f'`{field_name}` elements can be `{entity_type.__name__}`, got {type(element).__name__}; '
                        f'{element!r}; {field_name} = {field_value!r}.'
                    )
                
                validated[element.id] = element
        
        return validated
    
    return validator


def nullable_entity_dictionary_validator_factory(field_name, entity_type):
    """
    Returns a nullable entity dictionary validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    entity_type : `type<DiscordEntity>`
        The allowed entity type.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(field_value):
        """
        Validates the given nullable entity dictionary field.
        
        Parameters
        ----------
        field_value : `None`, `iterable` of `instance<entity_type>`, `dict` of (`int`, `entity_type`) items
            The field value to validate
        
        Returns
        -------
        entity_dictionary : `None`, `dict` of `instance<entity_type>`
        
        Raises
        ------
        TypeError
            - If `field_value`'s or it's elements type is incorrect.
        """
        nonlocal field_name
        nonlocal entity_type
        
        validated = None
        
        if field_value is not None:
            if isinstance(field_value, dict):
                iterator = iter(field_value.values())
            
            elif (getattr(field_value, '__iter__', None) is not None):
                iterator = iter(field_value)
            
            else:
                raise TypeError(
                    f'`{field_name}` can be `None`, `dict` of (`int`, `{entity_type.__name__}`) items, `iterable` of '
                    f'`{entity_type.__name__}` items, got {field_value.__class__.__name__}; {field_value!r}.'
                )
            
            for element in iterator:
                if not isinstance(element, entity_type):
                    raise TypeError(
                        f'`{field_name}` elements can be `{entity_type.__name__}`, got {type(element).__name__}; '
                        f'{element!r}; {field_name} = {field_value!r}.'
                    )
                
                if (validated is None):
                    validated = {}
                
                validated[element.id] = element
        
        return validated
    
    return validator


def duration_validator_factory(field_name):
    """
    Returns a duration validator.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(duration):
        """
        Duration validator returning when the duration ends.
        
        Parameters
        ----------
        duration : `TimeDelta`, `int`, `float`
            Duration in seconds or as a time delta.
        
        Returns
        -------
        until : `None | DateTime`
        """
        nonlocal field_name
        
        if isinstance(duration, int):
            if duration <= 0:
                return None
            
            return DateTime.now(TimeZone.utc) + TimeDelta(seconds = duration)
        
        if isinstance(duration, TimeDelta):
            if duration <= ZERO_TIMEDELTA:
                return None
            
            return DateTime.now(TimeZone.utc) + duration
        
        if isinstance(duration, float):
            if duration <= 0.0:
                return None
    
            return DateTime.now(TimeZone.utc) + TimeDelta(seconds = duration)
        
        raise TypeError(
            f'`{field_name}` can be `int`, `float`, `{TimeDelta.__name__}`, got '
            f'{type(duration).__name__}; {duration!r}.'
        )
    
    return validator


def nullable_bytes_validator_factory(field_name, length_min, length_max):
    """
    Returns a nullable bytes validator.
    
    Parameters
    ----------
    field_name : `bytes`
        The field's name.
    
    length_min : `int`
        The minimal allowed bytes length.
    
    length_max : `int`
        The maximal allowed bytes length.
    
    Returns
    -------
    validator : `FunctionType`
    """
    def validator(value):
        """
        Validates the given bytes value.
        
        > This function is generated.
        
        Parameters
        ----------
        value : `None | bytes`
            The bytes to validate.
        
        Returns
        -------
        value : `None | bytes`
                
        Raises
        ------
        TypeError
            - If `value` is not `None | bytes`.
        ValueError
            - If `value`'s length is out of the expected range.
        """
        nonlocal field_name
        nonlocal length_min
        nonlocal length_max
        
        if (value is not None):
            if not isinstance(value, bytes):
                raise TypeError(
                    f'`{field_name}` can be `None | bytes`, got {type(value).__name__}; {value!r}.'
                )
            
            value_length = len(value)
            if not value_length:
                value = None
            
            elif (value_length < length_min) or (value_length > length_max):
                raise ValueError(
                    f'`{field_name}\'s length can be in range [{length_min} : {length_max}], '
                    f'got {value_length}; {value!r}'
                )
        
        return value
    
    return validator
